@@ -1,6 +1,6 @@
 /*
    Copyright (c) 2006, 2013, Oracle and/or its affiliates.
-   Copyright (c) 2010, 2024, MariaDB
+   Copyright (c) 2010, 2017, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,14 +22,14 @@
 
 #include <welcome_copyright_notice.h> /* ORACLE_WELCOME_COPYRIGHT_NOTICE */
 
-#define VER "2.0"
+#define VER "2.1"
 
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
 
 #ifndef WEXITSTATUS
-# ifdef __WIN__
+# ifdef _WIN32
 #  define WEXITSTATUS(stat_val) (stat_val)
 # else
 #  define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
@@ -38,7 +38,7 @@
 
 static int phase = 0;
 static int info_file= -1;
-static const int phases_total = 7;
+static const int phases_total = 8;
 static char mysql_path[FN_REFLEN];
 static char mysqlcheck_path[FN_REFLEN];
 
@@ -87,10 +87,10 @@ static struct my_option my_long_options[]=
   {"basedir", 'b',
    "Not used by mysql_upgrade. Only for backward compatibility.",
    0, 0, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  {"character-sets-dir", 0,
+  {"character-sets-dir", OPT_CHARSETS_DIR,
    "Not used by mysql_upgrade. Only for backward compatibility.",
    0, 0, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0 },
-  {"compress", 0,
+  {"compress", OPT_COMPRESS,
    "Not used by mysql_upgrade. Only for backward compatibility.",
    &not_used, &not_used, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
   {"datadir", 'd',
@@ -103,12 +103,12 @@ static struct my_option my_long_options[]=
   {"debug", '#', "Output debug log.",
    0, 0, 0, GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
 #endif
-  {"debug-check", 0, "Check memory and open file usage at exit.",
+  {"debug-check", OPT_DEBUG_CHECK, "Check memory and open file usage at exit.",
    &debug_check_flag, &debug_check_flag,
    0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
   {"debug-info", 'T', "Print some debug info at exit.", &debug_info_flag,
    &debug_info_flag, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
-  {"default-character-set", 0,
+  {"default-character-set", OPT_DEFAULT_CHARSET,
    "Not used by mysql_upgrade. Only for backward compatibility.",
    0, 0, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"default_auth", OPT_DEFAULT_AUTH,
@@ -128,7 +128,7 @@ static struct my_option my_long_options[]=
    "Password to use when connecting to server. If password is not given,"
    " it's solicited on the tty.", &opt_password,&opt_password,
    0, GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
-#ifdef __WIN__
+#ifdef _WIN32
   {"pipe", 'W', "Use named pipes to connect to server.", 0, 0, 0,
    GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0},
 #endif
@@ -493,7 +493,7 @@ static int run_tool(char *tool_path, DYNAMIC_STRING *ds_res, ...)
 
   va_end(args);
 
-#ifdef __WIN__
+#ifdef _WIN32
   dynstr_append(&ds_cmdline, "\"");
 #endif
 
@@ -982,7 +982,7 @@ static my_bool is_mysql()
 
 static int run_mysqlcheck_views(void)
 {
-  const char *upgrade_views="--process-views=YES";
+  const char *upgrade_views="--process-views=UPGRADE";
   if (upgrade_from_mysql)
   {
     /*
@@ -1122,8 +1122,9 @@ static my_bool from_before_10_1()
 }
 
 
-static void uninstall_plugins(void)
+static int uninstall_plugins(void)
 {
+  verbose("Phase %d/%d: uninstalling plugins", ++phase, phases_total);
   if (ds_plugin_data_types.length)
   {
     char *plugins= ds_plugin_data_types.str;
@@ -1140,7 +1141,10 @@ static void uninstall_plugins(void)
       next= get_line(next);
     }
   }
+  return 0;
 }
+
+
 /**
   @brief     Install plugins for missing data types
   @details   Check for entries with "Unknown data type" in I_S.TABLES,
@@ -1191,6 +1195,8 @@ static int install_used_plugin_data_types(void)
   dynstr_free(&ds_result);
   return 0;
 }
+
+
 /*
   Check for entries with "Unknown storage engine" in I_S.TABLES,
   try to load plugins for these tables if available (MDEV-11942)
@@ -1241,6 +1247,7 @@ static int install_used_engines(void)
   return 0;
 }
 
+
 static int check_slave_repositories(void)
 {
   DYNAMIC_STRING ds_result;
@@ -1368,6 +1375,13 @@ static int run_sql_fix_privilege_tables(void)
 }
 
 
+static int flush_privileges(void)
+{
+  verbose("Phase %d/%d: Running 'FLUSH PRIVILEGES'", ++phase, phases_total);
+  return run_query("FLUSH PRIVILEGES", NULL, TRUE);
+}
+
+
 /**
   Check if the server version matches with the server version mysql_upgrade
   was compiled with.
@@ -1397,10 +1411,11 @@ static int check_version_match(void)
 
   if (calc_server_version((char *) version_str) != MYSQL_VERSION_ID)
   {
-    fprintf(stderr, "Error: Server version (%s) does not match with the "
-            "version of\nthe server (%s) with which this program was built/"
-            "distributed. You can\nuse --skip-version-check to skip this "
-            "check.\n", version_str, MYSQL_SERVER_VERSION);
+    fprintf(stderr, "Error: Server version (%s)\n"
+            "does not match the version of the server (%s)\n"
+            "with which this program was built/distributed. You can\n"
+            "use --skip-version-check to skip this check.\n",
+            version_str, MYSQL_SERVER_VERSION);
     return 1;
   }
   return 0;
@@ -1417,7 +1432,7 @@ int main(int argc, char **argv)
   load_defaults_or_exit("my", load_default_groups, &argc, &argv);
   defaults_argv= argv; /* Must be freed by 'free_defaults' */
 
-#if defined(__WIN__)
+#if defined(_WIN32)
   if (GetModuleFileName(NULL, self_name, FN_REFLEN) == 0)
 #endif
   {
@@ -1493,18 +1508,14 @@ int main(int argc, char **argv)
   if (run_mysqlcheck_upgrade(TRUE) ||
       install_used_engines() ||
       install_used_plugin_data_types() ||
-      run_mysqlcheck_views() ||
       run_sql_fix_privilege_tables() ||
+      run_mysqlcheck_views() ||
       run_mysqlcheck_fixnames() ||
       run_mysqlcheck_upgrade(FALSE) ||
-      check_slave_repositories())
+      check_slave_repositories() ||
+      uninstall_plugins() ||
+      flush_privileges())
     die("Upgrade failed" );
-
-  uninstall_plugins();
-  verbose("Phase %d/%d: Running 'FLUSH PRIVILEGES'", ++phase, phases_total);
-  if (run_query("FLUSH PRIVILEGES", NULL, TRUE))
-    die("Upgrade failed" );
-
   verbose("OK");
 
   /* Finish writing indicating upgrade has been performed */
