@@ -1,6 +1,6 @@
 /*
    Copyright (c) 2000, 2015, Oracle and/or its affiliates.
-   Copyright (c) 2011, 2024, MariaDB
+   Copyright (c) 2011, 2022, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -63,16 +63,17 @@ static uint     opt_mysql_port= 0, opt_protocol= 0;
 static char * opt_mysql_unix_port=0;
 static char *opt_plugin_dir= 0, *opt_default_auth= 0;
 static longlong opt_ignore_lines= -1;
+
 #include <sslopt-vars.h>
 
 static char **argv_to_free;
 
 static struct my_option my_long_options[] =
 {
-  {"character-sets-dir", 0,
+  {"character-sets-dir", OPT_CHARSETS_DIR,
    "Directory for character set files.", (char**) &charsets_dir,
    (char**) &charsets_dir, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  {"default-character-set", 0,
+  {"default-character-set", OPT_DEFAULT_CHARSET,
    "Set the default character set.", &default_charset,
    &default_charset, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"columns", 'c',
@@ -84,31 +85,31 @@ static struct my_option my_long_options[] =
    0, 0, 0},
   {"debug",'#', "Output debug log. Often this is 'd:t:o,filename'.", 0, 0, 0,
    GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
-  {"debug-check", 0, "Check memory and open file usage at exit.",
+  {"debug-check", OPT_DEBUG_CHECK, "Check memory and open file usage at exit.",
    &debug_check_flag, &debug_check_flag, 0,
    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
-  {"debug-info", 0, "Print some debug info at exit.",
+  {"debug-info", OPT_DEBUG_INFO, "Print some debug info at exit.",
    &debug_info_flag, &debug_info_flag,
    0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
-  {"default_auth", 0,
+  {"default_auth", OPT_DEFAULT_AUTH,
    "Default authentication client-side plugin to use.",
    &opt_default_auth, &opt_default_auth, 0,
    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"delete", 'd', "First delete all rows from table.", &opt_delete,
    &opt_delete, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
-  {"fields-terminated-by", 0,
+  {"fields-terminated-by", OPT_FTB,
    "Fields in the input file are terminated by the given string.", 
    &fields_terminated, &fields_terminated, 0, 
    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  {"fields-enclosed-by", 0,
+  {"fields-enclosed-by", OPT_ENC,
    "Fields in the import file are enclosed by the given character.", 
    &enclosed, &enclosed, 0, 
    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  {"fields-optionally-enclosed-by", 0,
+  {"fields-optionally-enclosed-by", OPT_O_ENC,
    "Fields in the input file are optionally enclosed by the given character.", 
    &opt_enclosed, &opt_enclosed, 0, 
    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  {"fields-escaped-by", 0, 
+  {"fields-escaped-by", OPT_ESC, 
    "Fields in the input file are escaped by the given character.",
    &escaped, &escaped, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0,
    0, 0},
@@ -125,10 +126,10 @@ static struct my_option my_long_options[] =
     "Disable foreign key checks while importing the data.",
     &ignore_foreign_keys, &ignore_foreign_keys, 0, GET_BOOL, NO_ARG,
     0, 0, 0, 0, 0, 0},
-  {"ignore-lines", 0, "Ignore first n lines of data infile.",
+  {"ignore-lines", OPT_IGN_LINES, "Ignore first n lines of data infile.",
    &opt_ignore_lines, &opt_ignore_lines, 0, GET_LL,
    REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  {"lines-terminated-by", 0, 
+  {"lines-terminated-by", OPT_LTB, 
    "Lines in the input file are terminated by the given string.",
    &lines_terminated, &lines_terminated, 0, GET_STR,
    REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
@@ -137,17 +138,17 @@ static struct my_option my_long_options[] =
   {"lock-tables", 'l', "Lock all tables for write (this disables threads).",
     &lock_tables, &lock_tables, 0, GET_BOOL, NO_ARG, 
     0, 0, 0, 0, 0, 0},
-  {"low-priority", 0,
+  {"low-priority", OPT_LOW_PRIORITY,
    "Use LOW_PRIORITY when updating the table.", &opt_low_priority,
    &opt_low_priority, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
   {"password", 'p',
    "Password to use when connecting to server. If password is not given it's asked from the tty.",
    0, 0, 0, GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
-#ifdef __WIN__
+#ifdef _WIN32
   {"pipe", 'W', "Use named pipes to connect to server.", 0, 0, 0, GET_NO_ARG,
    NO_ARG, 0, 0, 0, 0, 0, 0},
 #endif
-  {"plugin_dir", 0, "Directory for client-side plugins.",
+  {"plugin_dir", OPT_PLUGIN_DIR, "Directory for client-side plugins.",
    &opt_plugin_dir, &opt_plugin_dir, 0,
    GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"port", 'P', "Port number to use for connection or 0 for default to, in "
@@ -169,7 +170,7 @@ static struct my_option my_long_options[] =
    &opt_mysql_unix_port, &opt_mysql_unix_port, 0, GET_STR,
    REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
 #include <sslopt-longopts.h>
-  {"use-threads", 0,
+  {"use-threads", OPT_USE_THREADS,
    "Load files in parallel. The argument is the number "
    "of threads to use for loading data.",
    &opt_use_threads, &opt_use_threads, 0, 
@@ -222,7 +223,7 @@ file. The SQL command 'LOAD DATA INFILE' is used to import the rows.\n");
 
 static my_bool
 get_one_option(const struct my_option *opt, const char *argument,
-               const char *filename __attribute__((unused)))
+               const char *filename)
 {
   switch(opt->id) {
   case 'p':
@@ -246,7 +247,7 @@ get_one_option(const struct my_option *opt, const char *argument,
     else
       tty_password= 1;
     break;
-#ifdef __WIN__
+#ifdef _WIN32
   case 'W':
     opt_protocol = MYSQL_PROTOCOL_PIPE;
     opt_local_file=1;
@@ -260,6 +261,27 @@ get_one_option(const struct my_option *opt, const char *argument,
       exit(1);
     }
     break;
+  case 'P':
+    if (filename[0] == '\0')
+    {
+      /* Port given on command line, switch protocol to use TCP */
+      opt_protocol= MYSQL_PROTOCOL_TCP;
+    }
+    break;
+  case 'S':
+    if (filename[0] == '\0')
+    {
+      /*
+        Socket given on command line, switch protocol to use SOCKETSt
+        Except on Windows if 'protocol= pipe' has been provided in
+        the config file or command line.
+      */
+      if (opt_protocol != MYSQL_PROTOCOL_PIPE)
+      {
+        opt_protocol= MYSQL_PROTOCOL_SOCKET;
+      }
+    }
+    break;
   case '#':
     DBUG_PUSH(argument ? argument : "d:t:o");
     debug_check_flag= 1;
@@ -414,7 +436,7 @@ static void lock_table(MYSQL *mysql, int tablecount, char **raw_tablename)
     dynstr_append(&query, " WRITE,");
   }
   if (mysql_real_query(mysql, query.str, (ulong)query.length-1))
-    db_error(mysql); /* We shall continue here, if --force was given */
+    db_error(mysql); /* We shall countinue here, if --force was given */
 }
 
 
@@ -612,7 +634,7 @@ pthread_handler_t worker_thread(void *arg)
 
   if (mysql_query(mysql, "/*!40101 set @@character_set_database=binary */;"))
   {
-    db_error(mysql); /* We shall continue here, if --force was given */
+    db_error(mysql); /* We shall countinue here, if --force was given */
     goto error;
   }
 
@@ -643,6 +665,9 @@ int main(int argc, char **argv)
   MY_INIT(argv[0]);
   sf_leaking_memory=1; /* don't report memory leaks on early exits */
 
+  /* We need to know if protocol-related options originate from CLI args */
+  my_defaults_mark_files = TRUE;
+
   load_defaults_or_exit("my", load_default_groups, &argc, &argv);
   /* argv is changed in the program */
   argv_to_free= argv;
@@ -651,6 +676,7 @@ int main(int argc, char **argv)
     free_defaults(argv_to_free);
     return(1);
   }
+
   sf_leaking_memory=0; /* from now on we cleanup properly */
 
   if (opt_use_threads && !lock_tables)
@@ -736,12 +762,12 @@ int main(int argc, char **argv)
     if (!(mysql= db_connect(current_host,current_db,current_user,opt_password)))
     {
       free_defaults(argv_to_free);
-      return(1); /* purecov: dead code */
+      return(1); /* purecov: deadcode */
     }
 
     if (mysql_query(mysql, "/*!40101 set @@character_set_database=binary */;"))
     {
-      db_error(mysql); /* We shall continue here, if --force was given */
+      db_error(mysql); /* We shall countinue here, if --force was given */
       return(1);
     }
 
