@@ -23,10 +23,13 @@
 
 #include "my_decimal_limits.h"
 
-#define HOSTNAME_LENGTH 60
+#define HOSTNAME_LENGTH 255
+#define HOSTNAME_LENGTH_STR STRINGIFY_ARG(HOSTNAME_LENGTH)
 #define SYSTEM_CHARSET_MBMAXLEN 3
-#define NAME_CHAR_LEN	64U             /* Field/table name length */
+#define NAME_CHAR_LEN	64              /* Field/table name length */
 #define USERNAME_CHAR_LENGTH 128
+#define USERNAME_CHAR_LENGTH_STR STRINGIFY_ARG(USERNAME_CHAR_LENGTH)
+
 #define NAME_LEN                (NAME_CHAR_LEN*SYSTEM_CHARSET_MBMAXLEN)
 #define USERNAME_LENGTH         (USERNAME_CHAR_LENGTH*SYSTEM_CHARSET_MBMAXLEN)
 #define DEFINER_CHAR_LENGTH     (USERNAME_CHAR_LENGTH + HOSTNAME_LENGTH + 1)
@@ -71,6 +74,7 @@
 #define COLUMN_COMMENT_MAXLEN 1024
 #define INDEX_COMMENT_MAXLEN 1024
 #define TABLE_PARTITION_COMMENT_MAXLEN 1024
+#define DATABASE_COMMENT_MAXLEN 1024
 
 /*
   Maximum length of protocol packet.
@@ -92,10 +96,10 @@
 #define LOCAL_HOST_NAMEDPIPE "."
 
 
-#if defined(__WIN__) && !defined( _CUSTOMCONFIG_)
+#if defined(_WIN32) && !defined( _CUSTOMCONFIG_)
 #define MYSQL_NAMEDPIPE "MySQL"
 #define MYSQL_SERVICENAME "MySQL"
-#endif /* __WIN__ */
+#endif
 
 /*
   You should add new commands to the end of this list, otherwise old
@@ -120,7 +124,7 @@ enum enum_server_command
   COM_SLAVE_WORKER=251,
   COM_SLAVE_IO=252,
   COM_SLAVE_SQL=253,
-  COM_MULTI=254,
+  COM_RESERVED_1=254, /* Old COM_MULTI, now removed */
   /* Must be last */
   COM_END=255
 };
@@ -205,6 +209,8 @@ enum enum_indicator_type
                                                 itself supports it*/
 #define LONG_UNIQUE_HASH_FIELD       (1<< 30) /* This field will store hash for unique
                                                 column */
+#define FIELD_PART_OF_TMP_UNIQUE     (1<< 31) /* part of an unique constrain
+                                                for a tmporary table*/
 
 #define REFRESH_GRANT           (1ULL << 0)  /* Refresh grant tables */
 #define REFRESH_LOG             (1ULL << 1)  /* Start on new log file */
@@ -291,10 +297,17 @@ enum enum_indicator_type
 #define MARIADB_CLIENT_FLAGS_MASK 0xffffffff00000000ULL
 /* Client support progress indicator */
 #define MARIADB_CLIENT_PROGRESS (1ULL << 32)
-/* support COM_MULTI */
-#define MARIADB_CLIENT_COM_MULTI (1ULL << 33)
+
+/* Old COM_MULTI experiment (functionality removed).*/
+#define MARIADB_CLIENT_RESERVED_1 (1ULL << 33)
+
 /* support of array binding */
 #define MARIADB_CLIENT_STMT_BULK_OPERATIONS (1ULL << 34)
+/* support of extended metadata (e.g. type/format information) */
+#define MARIADB_CLIENT_EXTENDED_METADATA (1ULL << 35)
+
+/* Do not resend metadata for prepared statements, since 10.6*/
+#define MARIADB_CLIENT_CACHE_METADATA (1ULL << 36)
 
 #ifdef HAVE_COMPRESS
 #define CAN_CLIENT_COMPRESS CLIENT_COMPRESS
@@ -333,10 +346,10 @@ enum enum_indicator_type
                            CLIENT_SESSION_TRACK |\
                            CLIENT_DEPRECATE_EOF |\
                            CLIENT_CONNECT_ATTRS |\
-                           MARIADB_CLIENT_COM_MULTI |\
                            MARIADB_CLIENT_STMT_BULK_OPERATIONS |\
+                           MARIADB_CLIENT_EXTENDED_METADATA|\
+                           MARIADB_CLIENT_CACHE_METADATA |\
                            CLIENT_CAN_HANDLE_EXPIRED_PASSWORDS)
-
 /*
   Switch off the flags that are optional and depending on build flags
   If any of the optional flags is supported by the build it will be switched
@@ -345,6 +358,15 @@ enum enum_indicator_type
 #define CLIENT_BASIC_FLAGS ((CLIENT_ALL_FLAGS & ~CLIENT_SSL) \
                                                & ~CLIENT_COMPRESS)
 
+enum mariadb_field_attr_t
+{
+  MARIADB_FIELD_ATTR_DATA_TYPE_NAME= 0,
+  MARIADB_FIELD_ATTR_FORMAT_NAME= 1
+};
+
+#define MARIADB_FIELD_ATTR_LAST MARIADB_FIELD_ATTR_FORMAT_NAME
+
+
 /**
   Is raised when a multi-statement transaction
   has been started, either explicitly, by means
@@ -455,8 +477,7 @@ typedef struct st_net {
   char net_skip_rest_factor;
   my_bool thread_specific_malloc;
   unsigned char compress;
-  my_bool unused3; /* Please remove with the next incompatible ABI change. */
-  my_bool using_proxy_protocol;
+  my_bool pkt_nr_can_be_reset;
   /*
     Pointer to query object in query cache, do not equal NULL (0) for
     queries in cache that have not stored its results yet
@@ -599,6 +620,15 @@ enum enum_session_state_type
   SESSION_TRACK_GTIDS,
   SESSION_TRACK_TRANSACTION_CHARACTERISTICS,  /* Transaction chistics */
   SESSION_TRACK_TRANSACTION_STATE,            /* Transaction state */
+#ifdef USER_VAR_TRACKING
+  SESSION_TRACK_MYSQL_RESERVED1,
+  SESSION_TRACK_MYSQL_RESERVED2,
+  SESSION_TRACK_MYSQL_RESERVED3,
+  SESSION_TRACK_MYSQL_RESERVED4,
+  SESSION_TRACK_MYSQL_RESERVED5,
+  SESSION_TRACK_MYSQL_RESERVED6,
+  SESSION_TRACK_USER_VARIABLES,
+#endif // USER_VAR_TRACKING
   SESSION_TRACK_always_at_the_end             /* must be last */
 };
 
