@@ -102,7 +102,6 @@ static COND* make_cond(THD *thd, TABLE_LIST *tables, LEX_STRING *filter)
   if (!filter->str || !nrc)
     return 0;
 
-  nrc->init();
   nrc->resolve_in_table_list_only(tables);
   nrc->select_lex= tables->select_lex;
 
@@ -340,10 +339,6 @@ static int free(void *p)
     shutdown_plugin= true;
     mysql_cond_signal(&sleep_condition);
     mysql_mutex_unlock(&sleep_mutex);
-
-    for (uint i= 0; i < url_count; i++)
-      urls[i]->abort();
-
     pthread_join(sender_thread, NULL);
 
     mysql_mutex_destroy(&sleep_mutex);
