@@ -360,7 +360,7 @@ void field_unpack(String *to, Field *field, const uchar *rec, uint max_length,
   {
     if (field->is_null())
     {
-      to->append(STRING_WITH_LEN("NULL"));
+      to->append(NULL_clex_str);
       DBUG_VOID_RETURN;
     }
     CHARSET_INFO *cs= field->charset();
@@ -393,7 +393,7 @@ void field_unpack(String *to, Field *field, const uchar *rec, uint max_length,
     if (max_length < field->pack_length())
       tmp.length(MY_MIN(tmp.length(),max_length));
     ErrConvString err(&tmp);
-    to->append(err.ptr());
+    to->append(err.lex_cstring());
   }
   else
     to->append(STRING_WITH_LEN("???"));
@@ -434,7 +434,7 @@ void key_unpack(String *to, TABLE *table, KEY *key)
     {
       if (table->record[0][key_part->null_offset] & key_part->null_bit)
       {
-	to->append(STRING_WITH_LEN("NULL"));
+	to->append(NULL_clex_str);
 	continue;
       }
     }
