@@ -110,10 +110,12 @@ Range_rowid_filter_cost_info::set_adjusted_gain_param(double access_cost_factor)
 void Range_rowid_filter_cost_info::init(Rowid_filter_container_type cont_type,
                                         TABLE *tab, uint idx)
 {
+  DBUG_ASSERT(tab->opt_range_keys.is_set(idx));
+
   container_type= cont_type;
   table= tab;
   key_no= idx;
-  est_elements= (ulonglong) (table->quick_rows[key_no]);
+  est_elements= (ulonglong) table->opt_range[key_no].rows;
   b= build_cost(container_type);
   selectivity= est_elements/((double) table->stat_records());
   a= avg_access_and_eval_gain_per_row(container_type);
@@ -134,8 +136,9 @@ double
 Range_rowid_filter_cost_info::build_cost(Rowid_filter_container_type cont_type)
 {
   double cost= 0;
+  DBUG_ASSERT(table->opt_range_keys.is_set(key_no));
 
-  cost+= table->quick_index_only_costs[key_no];
+  cost+= table->opt_range[key_no].index_only_cost;
 
   switch (cont_type) {
 
@@ -345,7 +348,7 @@ void TABLE::init_cost_info_for_usable_range_rowid_filters(THD *thd)
   uint key_no;
   key_map usable_range_filter_keys;
   usable_range_filter_keys.clear_all();
-  key_map::Iterator it(quick_keys);
+  key_map::Iterator it(opt_range_keys);
 
   if (file->ha_table_flags() & HA_NON_COMPARABLE_ROWID)
     return;                                     // Cannot create filtering
@@ -360,9 +363,9 @@ void TABLE::init_cost_info_for_usable_range_rowid_filters(THD *thd)
   {
     if (!(file->index_flags(key_no, 0, 1) & HA_DO_RANGE_FILTER_PUSHDOWN))  // !1
       continue;
-    if (key_no == s->primary_key && file->primary_key_is_clustered())      // !2
+    if (file->is_clustering_key(key_no))                              // !2
       continue;
-   if (quick_rows[key_no] >
+   if (opt_range[key_no].rows >
        get_max_range_rowid_filter_elems_for_table(thd, this,
                                                   SORTED_ARRAY_CONTAINER)) // !3
       continue;
@@ -487,7 +490,7 @@ TABLE::best_range_rowid_filter_for_partial_join(uint access_key_no,
     clustered primary key it would, but the current InnoDB code does not
     allow it. Later this limitation will be lifted
   */
-  if (access_key_no == s->primary_key && file->primary_key_is_clustered())
+  if (file->is_clustering_key(access_key_no))
     return 0;
 
   Range_rowid_filter_cost_info *best_filter= 0;
