@@ -697,16 +697,13 @@ rpl_slave_state::record_gtid(THD *thd, const rpl_gtid *gtid, uint64 sub_id,
 
 #ifdef WITH_WSREP
   /*
-    We should replicate local gtid_slave_pos updates to other nodes if
-    wsrep gtid mode is set.
+    We should replicate local gtid_slave_pos updates to other nodes.
     In applier we should not append them to galera writeset.
   */
-  if (WSREP_ON_ && wsrep_gtid_mode && wsrep_thd_is_local(thd))
+  if (WSREP_ON_ && wsrep_thd_is_local(thd))
   {
     thd->wsrep_ignore_table= false;
     table->file->row_logging= 1; // replication requires binary logging
-    if (thd->wsrep_next_trx_id() == WSREP_UNDEFINED_TRX_ID)
-      thd->set_query_id(next_query_id());
     wsrep_start_trx_if_not_started(thd);
   }
   else
@@ -878,12 +875,10 @@ rpl_slave_state::gtid_delete_pending(THD *thd,
 
 #ifdef WITH_WSREP
   /*
-    We should replicate local gtid_slave_pos updates to other nodes if
-    wsrep gtid mode is set.
+    We should replicate local gtid_slave_pos updates to other nodes.
     In applier we should not append them to galera writeset.
   */
-  if (WSREP_ON_ && wsrep_gtid_mode &&
-      wsrep_thd_is_local(thd) &&
+  if (WSREP_ON_ && wsrep_thd_is_local(thd) &&
       thd->wsrep_cs().state() != wsrep::client_state::s_none)
   {
     if (thd->wsrep_trx().active() == false)
@@ -894,8 +889,7 @@ rpl_slave_state::gtid_delete_pending(THD *thd,
     }
     thd->wsrep_ignore_table= false;
   }
-  else
-    thd->wsrep_ignore_table= true;
+  thd->wsrep_ignore_table= true;
 #endif
 
   thd_saved_option= thd->variables.option_bits;
@@ -1077,13 +1071,13 @@ rpl_slave_state_tostring_helper(String *dest, const rpl_gtid *gtid, bool *first)
   if (*first)
     *first= false;
   else
-    if (dest->append(",",1))
+    if (dest->append(','))
       return true;
   return
     dest->append_ulonglong(gtid->domain_id) ||
-    dest->append("-",1) ||
+    dest->append('-') ||
     dest->append_ulonglong(gtid->server_id) ||
-    dest->append("-",1) ||
+    dest->append('-') ||
     dest->append_ulonglong(gtid->seq_no);
 }
 
@@ -1775,7 +1769,7 @@ rpl_binlog_state::alloc_element_nolock(const rpl_gtid *gtid)
 */
 bool
 rpl_binlog_state::check_strict_sequence(uint32 domain_id, uint32 server_id,
-                                        uint64 seq_no)
+                                        uint64 seq_no, bool no_error)
 {
   element *elem;
   bool res= 0;
@@ -1785,9 +1779,10 @@ rpl_binlog_state::check_strict_sequence(uint32 domain_id, uint32 server_id,
                                        (const uchar *)(&domain_id), 0)) &&
       elem->last_gtid && elem->last_gtid->seq_no >= seq_no)
   {
-    my_error(ER_GTID_STRICT_OUT_OF_ORDER, MYF(0), domain_id, server_id, seq_no,
-             elem->last_gtid->domain_id, elem->last_gtid->server_id,
-             elem->last_gtid->seq_no);
+    if (!no_error)
+      my_error(ER_GTID_STRICT_OUT_OF_ORDER, MYF(0), domain_id, server_id, seq_no,
+               elem->last_gtid->domain_id, elem->last_gtid->server_id,
+               elem->last_gtid->seq_no);
     res= 1;
   }
   mysql_mutex_unlock(&LOCK_binlog_state);
