@@ -25,8 +25,9 @@
 #include "my_stacktrace.h"
 #include <source_revision.h>
 
-#ifdef __WIN__
+#ifdef _WIN32
 #include <crtdbg.h>
+#include <direct.h>
 #define SIGNAL_FMT "exception 0x%x"
 #else
 #define SIGNAL_FMT "signal %d"
@@ -66,30 +67,30 @@ static inline void output_core_info()
                           (int) len, buff);
   }
 #ifdef __FreeBSD__
-  if ((fd= my_open("/proc/curproc/rlimit", O_RDONLY, MYF(0))) >= 0)
+  if ((fd= open("/proc/curproc/rlimit", O_RDONLY)) >= 0)
 #else
-  if ((fd= my_open("/proc/self/limits", O_RDONLY, MYF(0))) >= 0)
+  if ((fd= open("/proc/self/limits", O_RDONLY)) >= 0)
 #endif
   {
     my_safe_printf_stderr("Resource Limits:\n");
-    while ((len= my_read(fd, (uchar*)buff, sizeof(buff),  MYF(0))) > 0)
+    while ((len= read(fd, (uchar*)buff, sizeof(buff))) > 0)
     {
       my_write_stderr(buff, len);
     }
-    my_close(fd, MYF(0));
+    close(fd);
   }
 #ifdef __linux__
-  if ((fd= my_open("/proc/sys/kernel/core_pattern", O_RDONLY, MYF(0))) >= 0)
+  if ((fd= open("/proc/sys/kernel/core_pattern", O_RDONLY)) >= 0)
   {
-    len= my_read(fd, (uchar*)buff, sizeof(buff),  MYF(0));
+    len= read(fd, (uchar*)buff, sizeof(buff));
     my_safe_printf_stderr("Core pattern: %.*s\n", (int) len, buff);
-    my_close(fd, MYF(0));
+    close(fd);
   }
-  if ((fd= my_open("/proc/version", O_RDONLY, MYF(0))) >= 0)
+  if ((fd= open("/proc/version", O_RDONLY)) >= 0)
   {
-    len= my_read(fd, (uchar*)buff, sizeof(buff),  MYF(0));
+    len= read(fd, (uchar*)buff, sizeof(buff));
     my_safe_printf_stderr("Kernel version: %.*s\n", (int) len, buff);
-    my_close(fd, MYF(0));
+    close(fd);
   }
 #endif
 #elif defined(__APPLE__) || defined(__FreeBSD__)
@@ -103,11 +104,14 @@ static inline void output_core_info()
   {
     my_safe_printf_stderr("Kernel version: %.*s\n", (int) len, buff);
   }
-#else
+#elif defined(HAVE_GETCWD)
   char buff[80];
-  my_getwd(buff, sizeof(buff), 0);
-  my_safe_printf_stderr("Writing a core file at %s\n", buff);
-  fflush(stderr);
+
+  if (getcwd(buff, sizeof(buff)))
+  {
+    my_safe_printf_stderr("Writing a core file at %.*s\n", (int) sizeof(buff), buff);
+    fflush(stderr);
+  }
 #endif
 }
 
@@ -168,11 +172,8 @@ extern "C" sig_handler handle_fatal_signal(int sig)
   my_safe_printf_stderr("[ERROR] mysqld got " SIGNAL_FMT " ;\n",sig);
 
   my_safe_printf_stderr("%s",
-    "This could be because you hit a bug. It is also possible that this binary\n"
-    "or one of the libraries it was linked against is corrupt, improperly built,\n"
-    "or misconfigured. This error can also be caused by malfunctioning hardware.\n\n");
-
-  my_safe_printf_stderr("%s",
+                        "Sorry, we probably made a mistake, and this is a bug.\n\n"
+                        "Your assistance in bug reporting will enable us to fix this for the next release.\n"
                         "To report this bug, see https://mariadb.com/kb/en/reporting-bugs\n\n");
 
   my_safe_printf_stderr("%s",
@@ -305,7 +306,7 @@ extern "C" sig_handler handle_fatal_signal(int sig)
   }
   my_safe_printf_stderr("%s",
     "The manual page at "
-    "https://mariadb.com/kb/en/how-to-produce-a-full-stack-trace-for-mysqld/ contains\n"
+    "https://mariadb.com/kb/en/how-to-produce-a-full-stack-trace-for-mariadbd/ contains\n"
     "information that should help you find out what is causing the crash.\n");
 
 #endif /* HAVE_STACKTRACE */
@@ -358,7 +359,7 @@ extern "C" sig_handler handle_fatal_signal(int sig)
 #endif
 
 end:
-#ifndef __WIN__
+#ifndef _WIN32
   /*
      Quit, without running destructors (etc.)
      Use a signal, because the parent (systemd) can check that with WIFSIGNALED
