@@ -1,7 +1,7 @@
 /* -*- C++ -*- */
 /*
    Copyright (c) 2002, 2011, Oracle and/or its affiliates.
-   Copyright (c) 2020, MariaDB
+   Copyright (c) 2020, 2022, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -1081,8 +1081,9 @@ public:
     Query_arena(thd->lex->sphead->get_main_mem_root(), STMT_INITIALIZED_FOR_SP)
   { }
   ~sp_lex_cursor() { free_items(); }
-  void cleanup_stmt() { }
-  Query_arena *query_arena() { return this; }
+  virtual bool cleanup_stmt(bool /*restore_set_statement_vars*/) override
+  { return false; }
+  Query_arena *query_arena() override { return this; }
   bool validate()
   {
     DBUG_ASSERT(sql_command == SQLCOM_SELECT);
@@ -1125,7 +1126,7 @@ public:
   sp_instr(uint ip, sp_pcontext *ctx)
     :Query_arena(0, STMT_INITIALIZED_FOR_SP), marked(0), m_ip(ip), m_ctx(ctx)
 #ifdef PROTECT_STATEMENT_MEMROOT
-  , m_has_been_run(NON_RUN)
+  , m_has_been_run(false)
 #endif
   {}
 
@@ -1220,29 +1221,21 @@ public:
 #ifdef PROTECT_STATEMENT_MEMROOT
   bool has_been_run() const
   {
-    return m_has_been_run == RUN;
-  }
-
-  void mark_as_qc_used()
-  {
-    m_has_been_run= QC;
+    return m_has_been_run;
   }
 
   void mark_as_run()
   {
-    if (m_has_been_run == QC)
-      m_has_been_run= NON_RUN; // answer was from WC => not really executed
-    else
-      m_has_been_run= RUN;
+    m_has_been_run= true;
   }
 
   void mark_as_not_run()
   {
-    m_has_been_run= NON_RUN;
+    m_has_been_run= false;
   }
 
 private:
-  enum {NON_RUN, QC, RUN} m_has_been_run;
+  bool m_has_been_run;
 #endif
 }; // class sp_instr : public Sql_alloc
 
@@ -1879,23 +1872,24 @@ public:
 
   virtual ~sp_instr_cpush() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
   /**
     This call is used to cleanup the instruction when a sensitive
     cursor is closed. For now stored procedures always use materialized
     cursors and the call is not used.
   */
-  virtual void cleanup_stmt() { /* no op */ }
+  virtual bool cleanup_stmt(bool /*restore_set_statement_vars*/) override
+  { return false; }
 private:
 
   sp_lex_keeper m_lex_keeper;
   uint m_cursor;                /**< Frame offset (for debugging) */
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_cpush : public sp_instr
 
