@@ -37,6 +37,13 @@
 #endif
 
 const LEX_CSTRING msg_status= {STRING_WITH_LEN("status")};
+const LEX_CSTRING msg_repair= { STRING_WITH_LEN("repair") };
+const LEX_CSTRING msg_assign_to_keycache=
+{ STRING_WITH_LEN("assign_to_keycache") };
+const LEX_CSTRING msg_analyze= { STRING_WITH_LEN("analyze") };
+const LEX_CSTRING msg_check= { STRING_WITH_LEN("check") };
+const LEX_CSTRING msg_preload_keys= { STRING_WITH_LEN("preload_keys") };
+const LEX_CSTRING msg_optimize= { STRING_WITH_LEN("optimize") };
 
 /* Prepare, run and cleanup for mysql_recreate_table() */
 
@@ -78,15 +85,16 @@ static bool admin_recreate_table(THD *thd, TABLE_LIST *table_list,
 
 
 static int send_check_errmsg(THD *thd, TABLE_LIST* table,
-			     const char* operator_name, const char* errmsg)
+			     const LEX_CSTRING *operator_name,
+                             const char* errmsg)
 
 {
   Protocol *protocol= thd->protocol;
   protocol->prepare_for_resend();
   protocol->store(table->alias.str, table->alias.length, system_charset_info);
-  protocol->store((char*) operator_name, system_charset_info);
-  protocol->store(STRING_WITH_LEN("error"), system_charset_info);
-  protocol->store(errmsg, system_charset_info);
+  protocol->store(operator_name, system_charset_info);
+  protocol->store(&error_clex_str, system_charset_info);
+  protocol->store(errmsg, strlen(errmsg), system_charset_info);
   thd->clear_error();
   if (protocol->write())
     return -1;
@@ -155,7 +163,7 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
   */
   if (table->s->tmp_table)
   {
-    error= send_check_errmsg(thd, table_list, "repair",
+    error= send_check_errmsg(thd, table_list, &msg_repair,
 			     "Cannot repair temporary table from .frm file");
     goto end;
   }
@@ -173,8 +181,12 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
   if (table->s->frm_version < FRM_VER_TRUE_VARCHAR &&
       table->s->varchar_fields)
   {
-    error= send_check_errmsg(thd, table_list, "repair",
-                             "Failed repairing a very old .frm file as the data file format has changed between versions. Please dump the table in your old system with mysqldump and read it into this system with mysql or mysqlimport");
+    error= send_check_errmsg(thd, table_list, &msg_repair,
+                             "Failed repairing a very old .frm file as the "
+                             "data file format has changed between versions. "
+                             "Please dump the table in your old system with "
+                             "mysqldump and read it into this system with "
+                             "mysql or mysqlimport");
     goto end;
   }
 
@@ -232,12 +244,12 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
 
   if (my_rename(from, tmp, MYF(MY_WME)))
   {
-    error= send_check_errmsg(thd, table_list, "repair",
+    error= send_check_errmsg(thd, table_list, &msg_repair,
 			     "Failed renaming data file");
     goto end;
   }
   if (dd_recreate_table(thd, table_list->db.str, table_list->table_name.str))
-    create_error= send_check_errmsg(thd, table_list, "repair",
+    create_error= send_check_errmsg(thd, table_list, &msg_repair,
                                     "Failed generating table from .frm file");
   /*
     'FALSE' for 'using_transactions' means don't postpone
@@ -247,7 +259,7 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
   query_cache_invalidate3(thd, table_list, FALSE);
   if (mysql_file_rename(key_file_misc, tmp, from, MYF(MY_WME)))
   {
-    error= send_check_errmsg(thd, table_list, "repair",
+    error= send_check_errmsg(thd, table_list, &msg_repair,
 			     "Failed restoring .MYD file");
     goto end;
   }
@@ -269,7 +281,7 @@ static int prepare_for_repair(THD *thd, TABLE_LIST *table_list,
     */
     if (open_table(thd, table_list, &ot_ctx))
     {
-      error= send_check_errmsg(thd, table_list, "repair",
+      error= send_check_errmsg(thd, table_list, &msg_repair,
                                "Failed to open partially repaired table");
       goto end;
     }
@@ -478,6 +490,37 @@ static void send_read_only_warning(THD *thd, const LEX_CSTRING *msg_status,
 }
 
 
+/**
+  Collect field names of result set that will be sent to a client
+
+  @param      thd     Thread data object
+  @param[out] fields  List of fields whose metadata should be collected for
+                      sending to client
+*/
+
+void fill_check_table_metadata_fields(THD *thd, List<Item>* fields)
+{
+  Item *item;
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Table", NAME_CHAR_LEN * 2);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Op", 10);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Msg_type", 10);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Msg_text",
+                                              SQL_ADMIN_MSG_TEXT_SIZE);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+}
+
+
 /*
   RETURN VALUES
     FALSE Message sent to net (admin operation went ok)
@@ -486,7 +529,7 @@ static void send_read_only_warning(THD *thd, const LEX_CSTRING *msg_status,
 */
 static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
                               HA_CHECK_OPT* check_opt,
-                              const char *operator_name,
+                              const LEX_CSTRING *operator_name,
                               thr_lock_type lock_type,
                               bool org_open_for_modify,
                               bool repair_table_use_frm,
@@ -501,15 +544,12 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
 {
   TABLE_LIST *table;
   List<Item> field_list;
-  Item *item;
   Protocol *protocol= thd->protocol;
   LEX *lex= thd->lex;
-  int result_code= 0;
-  int compl_result_code= 0;
   bool need_repair_or_alter= 0;
   wait_for_commit* suspended_wfc;
   bool is_table_modified= false;
-
+  LEX_CUSTRING tabledef_version;
   DBUG_ENTER("mysql_admin_table");
   DBUG_PRINT("enter", ("extra_open_options: %u", extra_open_options));
 
@@ -525,21 +565,8 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   Disable_wsrep_on_guard wsrep_on_guard(thd, disable_wsrep_on);
 #endif /* WITH_WSREP */
 
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Table",
-                                         NAME_CHAR_LEN * 2), thd->mem_root);
-  item->maybe_null = 1;
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Op", 10), thd->mem_root);
-  item->maybe_null = 1;
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Msg_type", 10), thd->mem_root);
-  item->maybe_null = 1;
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Msg_text",
-                                         SQL_ADMIN_MSG_TEXT_SIZE),
-                       thd->mem_root);
-  item->maybe_null = 1;
+  fill_check_table_metadata_fields(thd, &field_list);
+
   if (protocol->send_result_set_metadata(&field_list,
                             Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF))
     DBUG_RETURN(TRUE);
@@ -565,12 +592,19 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   {
     char table_name_buff[SAFE_NAME_LEN*2+2];
     LEX_CSTRING table_name= { table_name_buff, 0};
+    char storage_engine_name[NAME_LEN];
+    bool storage_engine_partitioned= 0;
+    uchar tabledef_version_buff[MY_UUID_SIZE];
     const char *db= table->db.str;
     bool fatal_error=0;
-    bool open_error;
+    bool open_error= 0;
     bool collect_eis=  FALSE;
     bool open_for_modify= org_open_for_modify;
     Recreate_info recreate_info;
+    int compl_result_code, result_code;
+
+    compl_result_code= result_code= HA_ADMIN_FAILED;
+    storage_engine_name[0]= 0;                  // Marker that's not used
 
     DBUG_PRINT("admin", ("table: '%s'.'%s'", db, table->table_name.str));
     DEBUG_SYNC(thd, "admin_command_kill_before_modify");
@@ -587,7 +621,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
     */
     table->mdl_request.set_type(lex->sql_command == SQLCOM_REPAIR
                                 ? MDL_SHARED_NO_READ_WRITE
-                                : lock_type >= TL_WRITE_ALLOW_WRITE
+                                : lock_type >= TL_FIRST_WRITE
                                 ? MDL_SHARED_WRITE : MDL_SHARED_READ);
     if (thd->check_killed())
     {
@@ -663,7 +697,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             protocol->prepare_for_resend();
             protocol->store(&table_name, system_charset_info);
             protocol->store(operator_name, system_charset_info);
-            protocol->store(STRING_WITH_LEN("error"), system_charset_info);
+            protocol->store(&error_clex_str, system_charset_info);
             length= my_snprintf(buff, sizeof(buff),
                                 ER_THD(thd, ER_DROP_PARTITION_NON_EXISTENT),
                                 table_name.str);
@@ -747,7 +781,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
     {
       /* purecov: begin inspected */
       enum_sql_command save_sql_command= lex->sql_command;
-      LEX_CSTRING error_clex_str= { STRING_WITH_LEN("error") };
       DBUG_PRINT("admin", ("sending error message"));
       protocol->prepare_for_resend();
       protocol->store(&table_name, system_charset_info);
@@ -789,6 +822,16 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
         thd->close_unused_temporary_table_instances(table);
       else
       {
+        /* Store information about table for ddl log */
+        storage_engine_partitioned= table->table->file->partition_engine();
+        strmake(storage_engine_name, table->table->file->real_table_type(),
+                sizeof(storage_engine_name)-1);
+        tabledef_version.str= tabledef_version_buff;
+        if ((tabledef_version.length= table->table->s->tabledef_version.length))
+          memcpy((char*) tabledef_version.str,
+                 table->table->s->tabledef_version.str,
+                 MY_UUID_SIZE);
+
         if (wait_while_table_is_used(thd, table->table, HA_EXTRA_NOT_USED))
           goto err;
         DEBUG_SYNC(thd, "after_admin_flush");
@@ -863,20 +906,21 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       */
       collect_eis=
         (table->table->s->table_category == TABLE_CATEGORY_USER &&
-        !(lex->alter_info.partition_flags & ALTER_PARTITION_ADMIN) &&
+        !(lex->alter_info.flags & ALTER_PARTITION_ADMIN) &&
          (check_eits_collection_allowed(thd) ||
           lex->with_persistent_for_clause));
     }
 
     if (result_code == HA_ADMIN_OK)
     {    
-      DBUG_PRINT("admin", ("calling operator_func '%s'", operator_name));
+      DBUG_PRINT("admin", ("calling operator_func '%s'", operator_name->str));
       THD_STAGE_INFO(thd, stage_executing);
       result_code = (table->table->file->*operator_func)(thd, check_opt);
       THD_STAGE_INFO(thd, stage_sending_data);
       DBUG_PRINT("admin", ("operator_func returned: %d", result_code));
     }
 
+    /* Note: compl_result_code can be different from result_code here */
     if (compl_result_code == HA_ADMIN_OK && collect_eis)
     {
       if (result_code == HA_ERR_TABLE_READONLY)
@@ -917,13 +961,35 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
         Field **field_ptr= tab->field;
         if (!lex->column_list)
         {
+          /* Fields we have to read from the engine */
           bitmap_clear_all(tab->read_set);
+          /* Fields we want to have statistics for */
+          bitmap_clear_all(&tab->has_value_set);
+
           for (uint fields= 0; *field_ptr; field_ptr++, fields++)
           {
-            enum enum_field_types type= (*field_ptr)->type();
-            if (type < MYSQL_TYPE_MEDIUM_BLOB ||
+            Field *field= *field_ptr;
+            if (field->flags & LONG_UNIQUE_HASH_FIELD)
+            {
+              /*
+                No point in doing statistic for hash fields that should be
+                unique
+              */
+              continue;
+            }
+            /*
+              Note that type() always return MYSQL_TYPE_BLOB for
+              all blob types. Another function needs to be added
+              if we in the future want to distingush between blob
+              types here.
+            */
+            enum enum_field_types type= field->type();
+            if (type < MYSQL_TYPE_TINY_BLOB ||
                 type > MYSQL_TYPE_BLOB)
-              tab->field[fields]->register_field_in_read_map();
+            {
+              field->register_field_in_read_map();
+              bitmap_set_bit(&tab->has_value_set, field->field_index);
+            }
             else
               push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                                   ER_NO_EIS_FOR_FIELD,
@@ -937,9 +1003,15 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
           LEX_STRING *column_name;
           List_iterator_fast<LEX_STRING> it(*lex->column_list);
 
+          /* Fields we have to read from the engine */
           bitmap_clear_all(tab->read_set);
+          /* Fields we want to have statistics for */
+          bitmap_clear_all(&tab->has_value_set);
+
           while ((column_name= it++))
           {
+            Field *field;
+            enum enum_field_types type;
             if (tab->s->fieldnames.type_names == 0 ||
                 (pos= find_type(&tab->s->fieldnames, column_name->str,
                                 column_name->length, 1)) <= 0)
@@ -948,10 +1020,15 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
               break;
             }
             pos--;
-            enum enum_field_types type= tab->field[pos]->type();
-            if (type < MYSQL_TYPE_MEDIUM_BLOB ||
-                type > MYSQL_TYPE_BLOB)
-              tab->field[pos]->register_field_in_read_map();
+            field= tab->field[pos];
+            type= field->type();
+            if (!(field->flags & LONG_UNIQUE_HASH_FIELD) &&
+                (type < MYSQL_TYPE_TINY_BLOB ||
+                 type > MYSQL_TYPE_BLOB))
+            {
+              field->register_field_in_read_map();
+              bitmap_set_bit(&tab->has_value_set, field->field_index);
+            }
             else
               push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                                   ER_NO_EIS_FOR_FIELD,
@@ -981,11 +1058,14 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             tab->keys_in_use_for_query.set_bit(--pos);
           }
         }
+        /* Ensure that number of records are updated */
+        tab->file->info(HA_STATUS_VARIABLE);
         if (!(compl_result_code=
-              alloc_statistics_for_table(thd, table->table)) &&
+              alloc_statistics_for_table(thd, tab,
+                                         &tab->has_value_set)) &&
             !(compl_result_code=
-              collect_statistics_for_table(thd, table->table)))
-          compl_result_code= update_statistics_for_table(thd, table->table);
+              collect_statistics_for_table(thd, tab)))
+          compl_result_code= update_statistics_for_table(thd, tab);
       }
       else
         compl_result_code= HA_ADMIN_FAILED;
@@ -1022,7 +1102,7 @@ send_result:
     {
       protocol->prepare_for_resend();
       protocol->store(&table_name, system_charset_info);
-      protocol->store((char*) operator_name, system_charset_info);
+      protocol->store(operator_name, system_charset_info);
       protocol->store(warning_level_names[Sql_condition::WARN_LEVEL_WARN].str,
                       warning_level_names[Sql_condition::WARN_LEVEL_WARN].length,
                       system_charset_info);
@@ -1042,13 +1122,14 @@ send_result:
       const Sql_condition *err;
       while ((err= it++))
       {
+        const char *err_msg= err->get_message_text();
         protocol->prepare_for_resend();
         protocol->store(&table_name, system_charset_info);
-        protocol->store((char*) operator_name, system_charset_info);
+        protocol->store(operator_name, system_charset_info);
         protocol->store(warning_level_names[err->get_level()].str,
                         warning_level_names[err->get_level()].length,
                         system_charset_info);
-        protocol->store(err->get_message_text(), system_charset_info);
+        protocol->store(err_msg, strlen(err_msg), system_charset_info);
         if (protocol->write())
           goto err;
       }
@@ -1067,7 +1148,7 @@ send_result_message:
        char buf[MYSQL_ERRMSG_SIZE];
        size_t length=my_snprintf(buf, sizeof(buf),
                                  ER_THD(thd, ER_CHECK_NOT_IMPLEMENTED),
-                                 operator_name);
+                                 operator_name->str);
 	protocol->store(STRING_WITH_LEN("note"), system_charset_info);
 	protocol->store(buf, length, system_charset_info);
       }
@@ -1109,13 +1190,13 @@ send_result_message:
       break;
 
     case HA_ADMIN_CORRUPT:
-      protocol->store(STRING_WITH_LEN("error"), system_charset_info);
+      protocol->store(&error_clex_str, system_charset_info);
       protocol->store(STRING_WITH_LEN("Corrupt"), system_charset_info);
       fatal_error=1;
       break;
 
     case HA_ADMIN_INVALID:
-      protocol->store(STRING_WITH_LEN("error"), system_charset_info);
+      protocol->store(&error_clex_str, system_charset_info);
       protocol->store(STRING_WITH_LEN("Invalid argument"),
                       system_charset_info);
       break;
@@ -1198,8 +1279,8 @@ send_result_message:
           else
           {
             /* Hijack the row already in-progress. */
-            protocol->store(STRING_WITH_LEN("error"), system_charset_info);
-            protocol->store(err_msg, system_charset_info);
+            protocol->store(&error_clex_str, system_charset_info);
+            protocol->store(err_msg, strlen(err_msg), system_charset_info);
             if (protocol->write())
               goto err;
             /* Start off another row for HA_ADMIN_FAILED */
@@ -1235,7 +1316,7 @@ send_result_message:
       const char *what_to_upgrade= table->view ? "VIEW" :
           table->table->file->ha_table_flags() & HA_CAN_REPAIR ? "TABLE" : 0;
 
-      protocol->store(STRING_WITH_LEN("error"), system_charset_info);
+      protocol->store(&error_clex_str, system_charset_info);
       if (what_to_upgrade)
         length= my_snprintf(buf, sizeof(buf),
                             ER_THD(thd, ER_TABLE_NEEDS_UPGRADE),
@@ -1259,7 +1340,7 @@ send_result_message:
         size_t length=my_snprintf(buf, sizeof(buf),
                                 "Unknown - internal error %d during operation",
                                 result_code);
-        protocol->store(STRING_WITH_LEN("error"), system_charset_info);
+        protocol->store(&error_clex_str, system_charset_info);
         protocol->store(buf, length, system_charset_info);
         fatal_error=1;
         break;
@@ -1272,15 +1353,11 @@ send_result_message:
       with conflicting DMLs resulting in deadlock.
     */
     thd->transaction->stmt.mark_executed_table_admin_cmd();
+
     if (table->table && !table->view)
     {
-      /*
-        Don't skip flushing if we are collecting EITS statistics.
-      */
-      const bool skip_flush=
-        (operator_func == &handler::ha_analyze) && 
-        (table->table->file->ha_table_flags() & HA_ONLINE_ANALYZE) &&
-        !collect_eis;
+      /* Skip FLUSH TABLES if we are doing analyze */
+      const bool skip_flush= (operator_func == &handler::ha_analyze);
       if (table->table->s->tmp_table)
       {
         /*
@@ -1300,6 +1377,13 @@ send_result_message:
         table->table= 0;                        // For query cache
         query_cache_invalidate3(thd, table, 0);
       }
+      else if (collect_eis && skip_flush && compl_result_code == HA_ADMIN_OK)
+      {
+        TABLE_LIST *save_next_global= table->next_global;
+        table->next_global= 0;
+        read_statistics_for_tables(thd, table, true /* force_reload */);
+        table->next_global= save_next_global;
+      }
     }
     /* Error path, a admin command failed. */
     if (thd->transaction_rollback_request || fatal_error)
@@ -1318,6 +1402,22 @@ send_result_message:
       is_table_modified= true;
     }
     close_thread_tables(thd);
+
+    if (storage_engine_name[0])
+    {
+      /* Table was changed (repair, optimize or something similar) */
+      backup_log_info ddl_log;
+      bzero(&ddl_log, sizeof(ddl_log));
+      lex_string_set(&ddl_log.org_storage_engine_name,
+                     storage_engine_name);
+      ddl_log.query=            *operator_name;
+      ddl_log.org_partitioned=  storage_engine_partitioned;
+      ddl_log.org_database=     table->db;
+      ddl_log.org_table=        table->table_name;
+      ddl_log.org_table_id=     tabledef_version;
+      backup_log_ddl(&ddl_log);
+    }
+
     thd->release_transactional_locks();
 
     /*
@@ -1409,8 +1509,9 @@ bool mysql_assign_to_keycache(THD* thd, TABLE_LIST* tables,
   }
   check_opt.key_cache= key_cache;
   DBUG_RETURN(mysql_admin_table(thd, tables, &check_opt,
-				"assign_to_keycache", TL_READ_NO_INSERT, 0, 0,
-				0, 0, &handler::assign_to_keycache, 0, false));
+                                &msg_assign_to_keycache, TL_READ_NO_INSERT, 0,
+                                0, 0, 0,
+                                &handler::assign_to_keycache, 0, false));
 }
 
 
@@ -1436,8 +1537,9 @@ bool mysql_preload_keys(THD* thd, TABLE_LIST* tables)
     outdated information if parallel inserts into cache blocks happen.
   */
   DBUG_RETURN(mysql_admin_table(thd, tables, 0,
-				"preload_keys", TL_READ_NO_INSERT, 0, 0, 0, 0,
-				&handler::preload_keys, 0, false));
+                                &msg_preload_keys, TL_READ_NO_INSERT,
+                                0, 0, 0, 0,
+                                &handler::preload_keys, 0, false));
 }
 
 
@@ -1457,9 +1559,8 @@ bool Sql_cmd_analyze_table::execute(THD *thd)
 
   WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt,
-                         "analyze", lock_type, 1, 0, 0, 0,
+                         &msg_analyze, lock_type, 1, 0, 0, 0,
                          &handler::ha_analyze, 0, true);
-
   m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
@@ -1483,7 +1584,7 @@ bool Sql_cmd_check_table::execute(THD *thd)
                          TRUE, UINT_MAX, FALSE))
     goto error; /* purecov: inspected */
 
-  res= mysql_admin_table(thd, first_table, &m_lex->check_opt, "check",
+  res= mysql_admin_table(thd, first_table, &m_lex->check_opt, &msg_check,
                          lock_type, 0, 0, HA_OPEN_FOR_REPAIR, 0,
                          &handler::ha_check, &view_check, false);
 
@@ -1511,9 +1612,8 @@ bool Sql_cmd_optimize_table::execute(THD *thd)
   res= (specialflag & SPECIAL_NO_NEW_FUNC) ?
     mysql_recreate_table(thd, first_table, &recreate_info, true) :
     mysql_admin_table(thd, first_table, &m_lex->check_opt,
-                      "optimize", TL_WRITE, 1, 0, 0, 0,
+                      &msg_optimize, TL_WRITE, 1, 0, 0, 0,
                       &handler::ha_optimize, 0, true);
-
   m_lex->first_select_lex()->table_list.first= first_table;
   m_lex->query_tables= first_table;
 
@@ -1535,9 +1635,8 @@ bool Sql_cmd_repair_table::execute(THD *thd)
   if (check_table_access(thd, SELECT_ACL | INSERT_ACL, first_table,
                          FALSE, UINT_MAX, FALSE))
     goto error; /* purecov: inspected */
-
   WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, first_table);
-  res= mysql_admin_table(thd, first_table, &m_lex->check_opt, "repair",
+  res= mysql_admin_table(thd, first_table, &m_lex->check_opt, &msg_repair,
                          TL_WRITE, 1,
                          MY_TEST(m_lex->check_opt.sql_flags & TT_USEFRM),
                          HA_OPEN_FOR_REPAIR, &prepare_for_repair,
