@@ -130,6 +130,9 @@ TABLE *open_ltable(THD *thd, TABLE_LIST *table_list, thr_lock_type update,
 */
 #define MYSQL_OPEN_IGNORE_LOGGING_FORMAT        0x20000
 
+/* Don't use statistics tables */
+#define MYSQL_OPEN_IGNORE_ENGINE_STATS          0x40000
+
 /** Please refer to the internals manual. */
 #define MYSQL_OPEN_REOPEN  (MYSQL_OPEN_IGNORE_FLUSH |\
                             MYSQL_OPEN_IGNORE_GLOBAL_READ_LOCK |\
@@ -190,25 +193,26 @@ void unfix_fields(List<Item> &items);
 bool fill_record(THD * thd, TABLE *table_arg, List<Item> &fields,
                  List<Item> &values, bool ignore_errors, bool update);
 bool fill_record(THD *thd, TABLE *table, Field **field, List<Item> &values,
-                 bool ignore_errors, bool use_value,
-                 bool check_for_evaluability);
+                 bool ignore_errors, bool use_value);
 
 Field *
 find_field_in_tables(THD *thd, Item_ident *item,
                      TABLE_LIST *first_table, TABLE_LIST *last_table,
+                     ignored_tables_list_t ignored_tables,
                      Item **ref, find_item_error_report_type report_error,
                      bool check_privileges, bool register_tree_change);
 Field *
 find_field_in_table_ref(THD *thd, TABLE_LIST *table_list,
                         const char *name, size_t length,
                         const char *item_name, const char *db_name,
-                        const char *table_name, Item **ref,
-                        bool check_privileges, bool allow_rowid,
-                        uint *cached_field_index_ptr,
+                        const char *table_name,
+                        ignored_tables_list_t ignored_tables,
+                        Item **ref, bool check_privileges, bool allow_rowid,
+                        field_index_t *cached_field_index_ptr,
                         bool register_tree_change, TABLE_LIST **actual_table);
 Field *
 find_field_in_table(THD *thd, TABLE *table, const char *name, size_t length,
-                    bool allow_rowid, uint *cached_field_index_ptr);
+                    bool allow_rowid, field_index_t *cached_field_index_ptr);
 Field *
 find_field_in_table_sef(TABLE *table, const char *name);
 Item ** find_item_in_list(Item *item, List<Item> &items, uint *counter,
@@ -309,8 +313,7 @@ bool flush_tables(THD *thd, flush_tables_type flag);
 void close_all_tables_for_name(THD *thd, TABLE_SHARE *share,
                                ha_extra_function extra,
                                TABLE *skip_table);
-OPEN_TABLE_LIST *list_open_tables(THD *thd, const LEX_CSTRING &db,
-                                  const char *wild);
+OPEN_TABLE_LIST *list_open_tables(THD *thd, const char *db, const char *wild);
 bool tdc_open_view(THD *thd, TABLE_LIST *table_list, uint flags);
 
 TABLE *find_table_for_mdl_upgrade(THD *thd, const char *db,
@@ -419,13 +422,13 @@ public:
 class DML_prelocking_strategy : public Prelocking_strategy
 {
 public:
-  bool handle_routine(THD *thd, Query_tables_list *prelocking_ctx,
-                      Sroutine_hash_entry *rt, sp_head *sp,
-                      bool *need_prelocking) override;
-  bool handle_table(THD *thd, Query_tables_list *prelocking_ctx,
-                    TABLE_LIST *table_list, bool *need_prelocking) override;
-  bool handle_view(THD *thd, Query_tables_list *prelocking_ctx,
-                   TABLE_LIST *table_list, bool *need_prelocking) override;
+  virtual bool handle_routine(THD *thd, Query_tables_list *prelocking_ctx,
+                              Sroutine_hash_entry *rt, sp_head *sp,
+                              bool *need_prelocking);
+  virtual bool handle_table(THD *thd, Query_tables_list *prelocking_ctx,
+                            TABLE_LIST *table_list, bool *need_prelocking);
+  virtual bool handle_view(THD *thd, Query_tables_list *prelocking_ctx,
+                           TABLE_LIST *table_list, bool *need_prelocking);
 };
 
 
@@ -436,8 +439,8 @@ public:
 
 class Lock_tables_prelocking_strategy : public DML_prelocking_strategy
 {
-  bool handle_table(THD *thd, Query_tables_list *prelocking_ctx,
-                    TABLE_LIST *table_list, bool *need_prelocking) override;
+  virtual bool handle_table(THD *thd, Query_tables_list *prelocking_ctx,
+                            TABLE_LIST *table_list, bool *need_prelocking);
 };
 
 
@@ -452,13 +455,13 @@ class Lock_tables_prelocking_strategy : public DML_prelocking_strategy
 class Alter_table_prelocking_strategy : public Prelocking_strategy
 {
 public:
-  bool handle_routine(THD *thd, Query_tables_list *prelocking_ctx,
-                      Sroutine_hash_entry *rt, sp_head *sp,
-                      bool *need_prelocking) override;
-  bool handle_table(THD *thd, Query_tables_list *prelocking_ctx,
-                    TABLE_LIST *table_list, bool *need_prelocking) override;
-  bool handle_view(THD *thd, Query_tables_list *prelocking_ctx,
-                   TABLE_LIST *table_list, bool *need_prelocking) override;
+  virtual bool handle_routine(THD *thd, Query_tables_list *prelocking_ctx,
+                              Sroutine_hash_entry *rt, sp_head *sp,
+                              bool *need_prelocking);
+  virtual bool handle_table(THD *thd, Query_tables_list *prelocking_ctx,
+                            TABLE_LIST *table_list, bool *need_prelocking);
+  virtual bool handle_view(THD *thd, Query_tables_list *prelocking_ctx,
+                           TABLE_LIST *table_list, bool *need_prelocking);
 };
 
 
@@ -518,8 +521,6 @@ bool extend_table_list(THD *thd, TABLE_LIST *tables,
                        Prelocking_strategy *prelocking_strategy,
                        bool has_prelocking_list);
 
-void promote_select_describe_flag_if_needed(LEX *lex);
-
 /**
   A context of open_tables() function, used to recover
   from a failed open_table() or open_routine() attempt.
@@ -642,7 +643,7 @@ class No_such_table_error_handler : public Internal_error_handler
 {
 public:
   No_such_table_error_handler()
-    : m_handled_errors(0), m_unhandled_errors(0)
+    : m_handled_errors(0), m_unhandled_errors(0), first_error(0)
   {}
 
   bool handle_condition(THD *thd,
@@ -650,18 +651,18 @@ public:
                         const char* sqlstate,
                         Sql_condition::enum_warning_level *level,
                         const char* msg,
-                        Sql_condition ** cond_hdl) override;
+                        Sql_condition ** cond_hdl);
 
   /**
     Returns TRUE if one or more ER_NO_SUCH_TABLE errors have been
     trapped and no other errors have been seen. FALSE otherwise.
   */
   bool safely_trapped_errors();
+  uint got_error() { return first_error; }
 
 private:
   int m_handled_errors;
   int m_unhandled_errors;
+  uint first_error;
 };
-
-
 #endif /* SQL_BASE_INCLUDED */
