@@ -26,11 +26,12 @@
 #include "lock.h"                        // lock_schema_name
 #include "sql_table.h"                   // build_table_filename,
                                          // filename_to_tablename
+                                         // validate_comment_length
 #include "sql_rename.h"                  // mysql_rename_tables
 #include "sql_acl.h"                     // SELECT_ACL, DB_ACLS,
                                          // acl_get, check_grant_db
 #include "log_event.h"                   // Query_log_event
-#include "sql_base.h"                    // lock_table_names, tdc_remove_table
+#include "sql_base.h"                    // lock_table_names
 #include "sql_handler.h"                 // mysql_ha_rm_tables
 #include "sql_class.h"
 #include <mysys_err.h>
@@ -39,13 +40,14 @@
 #include "events.h"
 #include "sql_handler.h"
 #include "sql_statistics.h"
+#include "ddl_log.h"                     // ddl_log functions
 #include <my_dir.h>
 #include <m_ctype.h>
 #include "log.h"
-#ifdef __WIN__
+#ifdef _WIN32
 #include <direct.h>
 #endif
-#include "debug_sync.h"
+#include "debug.h"                       // debug_crash_here
 
 #define MAX_DROP_TABLE_Q_LEN      1024
 
@@ -57,10 +59,10 @@ static bool find_db_tables_and_rm_known_files(THD *, MY_DIR *, const char *,
                                               const char *, TABLE_LIST **);
 
 long mysql_rm_arc_files(THD *thd, MY_DIR *dirp, const char *org_path);
-static my_bool rm_dir_w_symlink(const char *org_path, my_bool send_error);
+my_bool rm_dir_w_symlink(const char *org_path, my_bool send_error);
 static void mysql_change_db_impl(THD *thd,
                                  LEX_CSTRING *new_db_name,
-                                 ulong new_db_access,
+                                 privilege_t new_db_access,
                                  CHARSET_INFO *new_db_charset);
 static bool mysql_rm_db_internal(THD *thd, const LEX_CSTRING *db,
                                  bool if_exists, bool silent);
@@ -77,6 +79,7 @@ typedef struct my_dbopt_st
   char *name;			/* Database name                  */
   uint name_length;		/* Database length name           */
   CHARSET_INFO *charset;	/* Database default character set */
+  LEX_STRING comment;           /* Database comment               */
 } my_dbopt_t;
 
 
@@ -102,8 +105,137 @@ cmp_db_names(LEX_CSTRING *db1_name, const LEX_CSTRING *db2_name)
                          db1_name->str, db2_name->str) == 0));
 }
 
+#ifdef HAVE_PSI_INTERFACE
+static PSI_rwlock_key key_rwlock_LOCK_dboptions;
+static PSI_rwlock_key key_rwlock_LOCK_dbnames;
+static PSI_rwlock_key key_rwlock_LOCK_rmdir;
+
+static PSI_rwlock_info all_database_names_rwlocks[]= {
+    {&key_rwlock_LOCK_dboptions, "LOCK_dboptions", PSI_FLAG_GLOBAL},
+    {&key_rwlock_LOCK_dbnames, "LOCK_dbnames", PSI_FLAG_GLOBAL},
+    {&key_rwlock_LOCK_rmdir, "LOCK_rmdir",PSI_FLAG_GLOBAL},
+};
+
+static void init_database_names_psi_keys(void)
+{
+  const char *category= "sql";
+  int count;
+
+  if (PSI_server == NULL)
+    return;
+
+  count= array_elements(all_database_names_rwlocks);
+  PSI_server->register_rwlock(category, all_database_names_rwlocks, count);
+}
+#endif
+
+static mysql_rwlock_t rmdir_lock;
 
 /*
+  Cache of C strings for existing database names.
+
+  The only use of it is to avoid repeated expensive
+  my_access() calls.
+
+  Provided operations are lookup, insert (after successfull my_access())
+  and clear (this is called whenever rmdir is called).
+*/
+struct dbname_cache_t
+{
+private:
+  Hash_set<LEX_STRING> m_set;
+  mysql_rwlock_t m_lock;
+
+  static uchar *get_key(const LEX_STRING *ls, size_t *sz, my_bool)
+  {
+    *sz= ls->length;
+    return (uchar *) ls->str;
+  }
+
+public:
+  dbname_cache_t()
+      : m_set(key_memory_dbnames_cache, table_alias_charset, 10, 0,
+              sizeof(char *), (my_hash_get_key) get_key, my_free, 0)
+  {
+    mysql_rwlock_init(key_rwlock_LOCK_dbnames, &m_lock);
+  }
+
+  bool contains(const char *s)
+  {
+    auto sz= strlen(s);
+    mysql_rwlock_rdlock(&m_lock);
+    bool ret= m_set.find(s, sz) != 0;
+    mysql_rwlock_unlock(&m_lock);
+    return ret;
+  }
+
+  void insert(const char *s)
+  {
+    auto len= strlen(s);
+    auto ls= (LEX_STRING *) my_malloc(key_memory_dbnames_cache,
+                                      sizeof(LEX_STRING) + strlen(s) + 1, 0);
+
+    if (!ls)
+      return;
+
+    ls->length= len;
+    ls->str= (char *) (ls + 1);
+
+    memcpy(ls->str, s, len + 1);
+    mysql_rwlock_wrlock(&m_lock);
+    bool found= m_set.find(s, len) != 0;
+    if (!found)
+      m_set.insert(ls);
+    mysql_rwlock_unlock(&m_lock);
+    if (found)
+      my_free(ls);
+  }
+
+  void clear()
+  {
+    mysql_rwlock_wrlock(&m_lock);
+    m_set.clear();
+    mysql_rwlock_unlock(&m_lock);
+  }
+
+  ~dbname_cache_t()
+  {
+      mysql_rwlock_destroy(&m_lock);
+  }
+};
+
+static dbname_cache_t* dbname_cache;
+
+static void dbname_cache_init()
+{
+  static MY_ALIGNED(16) char buf[sizeof(dbname_cache_t)];
+  DBUG_ASSERT(!dbname_cache);
+  dbname_cache= new (buf) dbname_cache_t;
+  mysql_rwlock_init(key_rwlock_LOCK_rmdir, &rmdir_lock);
+}
+
+static void dbname_cache_destroy()
+{
+  if (!dbname_cache)
+    return;
+
+  dbname_cache->~dbname_cache_t();
+  dbname_cache= 0;
+  mysql_rwlock_destroy(&rmdir_lock);
+}
+
+static int my_rmdir(const char *dir)
+{
+  auto ret= rmdir(dir);
+  if (ret)
+    return ret;
+  mysql_rwlock_wrlock(&rmdir_lock);
+  dbname_cache->clear();
+  mysql_rwlock_unlock(&rmdir_lock);
+  return 0;
+}
+
+  /*
   Function we use in the creation of our hash to get key.
 */
 
@@ -129,7 +261,7 @@ static inline int write_to_binlog(THD *thd, const char *query, size_t q_len,
   qinfo.db= db;
   qinfo.db_len= (uint32)db_len;
   return mysql_bin_log.write(&qinfo);
-}  
+}
 
 
 /*
@@ -143,26 +275,7 @@ void free_dbopt(void *dbopt)
   my_free(dbopt);
 }
 
-#ifdef HAVE_PSI_INTERFACE
-static PSI_rwlock_key key_rwlock_LOCK_dboptions;
 
-static PSI_rwlock_info all_database_names_rwlocks[]=
-{
-  { &key_rwlock_LOCK_dboptions, "LOCK_dboptions", PSI_FLAG_GLOBAL}
-};
-
-static void init_database_names_psi_keys(void)
-{
-  const char* category= "sql";
-  int count;
-
-  if (PSI_server == NULL)
-    return;
-
-  count= array_elements(all_database_names_rwlocks);
-  PSI_server->register_rwlock(category, all_database_names_rwlocks, count);
-}
-#endif
 
 /**
   Initialize database option cache.
@@ -184,10 +297,11 @@ bool my_dboptions_cache_init(void)
   if (!dboptions_init)
   {
     dboptions_init= 1;
-    error= my_hash_init(&dboptions, table_alias_charset,
-                        32, 0, 0, (my_hash_get_key) dboptions_get_key,
-                        free_dbopt,0);
+    error= my_hash_init(key_memory_dboptions_hash, &dboptions,
+                        table_alias_charset, 32, 0, 0, (my_hash_get_key)
+                        dboptions_get_key, free_dbopt, 0);
   }
+  dbname_cache_init();
   return error;
 }
 
@@ -203,6 +317,7 @@ void my_dboptions_cache_free(void)
   {
     dboptions_init= 0;
     my_hash_free(&dboptions);
+    dbname_cache_destroy();
     mysql_rwlock_destroy(&LOCK_dboptions);
   }
 }
@@ -216,9 +331,8 @@ void my_dbopt_cleanup(void)
 {
   mysql_rwlock_wrlock(&LOCK_dboptions);
   my_hash_free(&dboptions);
-  my_hash_init(&dboptions, table_alias_charset,
-               32, 0, 0, (my_hash_get_key) dboptions_get_key,
-               free_dbopt,0);
+  my_hash_init(key_memory_dboptions_hash, &dboptions, table_alias_charset, 32,
+               0, 0, (my_hash_get_key) dboptions_get_key, free_dbopt, 0);
   mysql_rwlock_unlock(&LOCK_dboptions);
 }
 
@@ -235,7 +349,8 @@ void my_dbopt_cleanup(void)
     1 on error.
 */
 
-static my_bool get_dbopt(const char *dbname, Schema_specification_st *create)
+static my_bool get_dbopt(THD *thd, const char *dbname,
+                         Schema_specification_st *create)
 {
   my_dbopt_t *opt;
   uint length;
@@ -247,6 +362,11 @@ static my_bool get_dbopt(const char *dbname, Schema_specification_st *create)
   if ((opt= (my_dbopt_t*) my_hash_search(&dboptions, (uchar*) dbname, length)))
   {
     create->default_table_charset= opt->charset;
+    if (opt->comment.length)
+    {
+      create->schema_comment= thd->make_clex_string(opt->comment.str,
+                                                    opt->comment.length);
+    }
     error= 0;
   }
   mysql_rwlock_unlock(&LOCK_dboptions);
@@ -274,15 +394,17 @@ static my_bool put_dbopt(const char *dbname, Schema_specification_st *create)
   DBUG_ENTER("put_dbopt");
 
   length= (uint) strlen(dbname);
-  
+
   mysql_rwlock_wrlock(&LOCK_dboptions);
   if (!(opt= (my_dbopt_t*) my_hash_search(&dboptions, (uchar*) dbname,
                                           length)))
   { 
     /* Options are not in the hash, insert them */
     char *tmp_name;
-    if (!my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
+    char *tmp_comment= NULL;
+    if (!my_multi_malloc(key_memory_dboptions_hash, MYF(MY_WME | MY_ZEROFILL),
                          &opt, (uint) sizeof(*opt), &tmp_name, (uint) length+1,
+                         &tmp_comment, (uint) DATABASE_COMMENT_MAXLEN+1,
                          NullS))
     {
       error= 1;
@@ -292,7 +414,7 @@ static my_bool put_dbopt(const char *dbname, Schema_specification_st *create)
     opt->name= tmp_name;
     strmov(opt->name, dbname);
     opt->name_length= length;
-    
+    opt->comment.str= tmp_comment;
     if (unlikely((error= my_hash_insert(&dboptions, (uchar*) opt))))
     {
       my_free(opt);
@@ -303,6 +425,12 @@ static my_bool put_dbopt(const char *dbname, Schema_specification_st *create)
   /* Update / write options in hash */
   opt->charset= create->default_table_charset;
 
+  if (create->schema_comment)
+  {
+    strmov(opt->comment.str, create->schema_comment->str);
+    opt->comment.length= create->schema_comment->length;
+  }
+
 end:
   mysql_rwlock_unlock(&LOCK_dboptions);
   DBUG_RETURN(error);
@@ -328,7 +456,8 @@ static void del_dbopt(const char *path)
   Create database options file:
 
   DESCRIPTION
-    Currently database default charset is only stored there.
+    Currently database default charset, default collation
+    and comment are stored there.
 
   RETURN VALUES
   0	ok
@@ -339,9 +468,34 @@ static bool write_db_opt(THD *thd, const char *path,
                          Schema_specification_st *create)
 {
   File file;
-  char buf[256]; // Should be enough for one option
+  char buf[256+DATABASE_COMMENT_MAXLEN];
   bool error=1;
 
+  if (create->schema_comment)
+  {
+    if (validate_comment_length(thd, create->schema_comment,
+                                DATABASE_COMMENT_MAXLEN,
+                                ER_TOO_LONG_DATABASE_COMMENT,
+                                thd->lex->name.str))
+      return error;
+  }
+
+  if (thd->lex->sql_command == SQLCOM_ALTER_DB &&
+      (!create->schema_comment || !create->default_table_charset))
+  {
+    /* Use existing values of schema_comment and charset for
+       ALTER DATABASE queries */
+    Schema_specification_st tmp;
+    tmp.init();
+    load_db_opt(thd, path, &tmp);
+
+    if (!create->schema_comment)
+      create->schema_comment= tmp.schema_comment;
+
+    if (!create->default_table_charset)
+      create->default_table_charset= tmp.default_table_charset;
+  }
+
   if (!create->default_table_charset)
     create->default_table_charset= thd->variables.collation_server;
 
@@ -353,11 +507,16 @@ static bool write_db_opt(THD *thd, const char *path,
   {
     ulong length;
     length= (ulong) (strxnmov(buf, sizeof(buf)-1, "default-character-set=",
-                              create->default_table_charset->csname,
+                              create->default_table_charset->cs_name.str,
                               "\ndefault-collation=",
-                              create->default_table_charset->name,
+                              create->default_table_charset->coll_name.str,
                               "\n", NullS) - buf);
 
+    if (create->schema_comment)
+      length= (ulong) (strxnmov(buf+length, sizeof(buf)-1-length,
+                                "comment=", create->schema_comment->str,
+                                "\n", NullS) - buf);
+
     /* Error is written by mysql_file_write */
     if (!mysql_file_write(file, (uchar*) buf, length, MYF(MY_NABP+MY_WME)))
       error=0;
@@ -385,16 +544,17 @@ static bool write_db_opt(THD *thd, const char *path,
 bool load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
 {
   File file;
-  char buf[256];
+  char buf[256+DATABASE_COMMENT_MAXLEN];
   DBUG_ENTER("load_db_opt");
   bool error=1;
   size_t nbytes;
+  myf utf8_flag= thd->get_utf8_flag();
 
   bzero((char*) create,sizeof(*create));
   create->default_table_charset= thd->variables.collation_server;
 
   /* Check if options for this database are already in the hash */
-  if (!get_dbopt(path, create))
+  if (!get_dbopt(thd, path, create))
     DBUG_RETURN(0);
 
   /* Otherwise, load options from the .opt file */
@@ -425,9 +585,9 @@ bool load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
            default-collation commands.
         */
         if (!(create->default_table_charset=
-        get_charset_by_csname(pos+1, MY_CS_PRIMARY, MYF(0))) &&
+        get_charset_by_csname(pos+1, MY_CS_PRIMARY, MYF(utf8_flag))) &&
             !(create->default_table_charset=
-              get_charset_by_name(pos+1, MYF(0))))
+              get_charset_by_name(pos+1, MYF(utf8_flag))))
         {
           sql_print_error("Error while loading database options: '%s':",path);
           sql_print_error(ER_THD(thd, ER_UNKNOWN_CHARACTER_SET),pos+1);
@@ -436,14 +596,15 @@ bool load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
       }
       else if (!strncmp(buf,"default-collation", (pos-buf)))
       {
-        if (!(create->default_table_charset= get_charset_by_name(pos+1,
-                                                           MYF(0))))
+        if (!(create->default_table_charset= get_charset_by_name(pos+1, MYF(utf8_flag))))
         {
           sql_print_error("Error while loading database options: '%s':",path);
           sql_print_error(ER_THD(thd, ER_UNKNOWN_COLLATION),pos+1);
           create->default_table_charset= default_charset_info;
         }
       }
+      else if (!strncmp(buf, "comment", (pos-buf)))
+        create->schema_comment= thd->make_clex_string(pos+1, strlen(pos+1));
     }
   }
   /*
@@ -544,7 +705,7 @@ CHARSET_INFO *get_default_db_collation(THD *thd, const char *db_name)
   Create a database
 
   SYNOPSIS
-  mysql_create_db_iternal()
+  mysql_create_db_internal()
   thd		Thread handler
   db		Name of database to create
 		Function assumes that this is already validated.
@@ -628,7 +789,6 @@ mysql_create_db_internal(THD *thd, const LEX_CSTRING *db,
     DBUG_RETURN(-1);
   }
 
-
   if (my_mkdir(path, 0777, MYF(0)) < 0)
   {
     my_error(ER_CANT_CREATE_DB, MYF(0), db->str, my_errno);
@@ -644,7 +804,7 @@ mysql_create_db_internal(THD *thd, const LEX_CSTRING *db,
       Restore things to beginning.
     */
     path[path_len]= 0;
-    if (rmdir(path) >= 0)
+    if (my_rmdir(path) >= 0)
       DBUG_RETURN(-1);
     /*
       We come here when we managed to create the database, but not the option
@@ -654,6 +814,14 @@ mysql_create_db_internal(THD *thd, const LEX_CSTRING *db,
     thd->clear_error();
   }
 
+  /* Log command to ddl log */
+  backup_log_info ddl_log;
+  bzero(&ddl_log, sizeof(ddl_log));
+  ddl_log.query=                   { C_STRING_WITH_LEN("CREATE") };
+  ddl_log.org_storage_engine_name= { C_STRING_WITH_LEN("DATABASE") };
+  ddl_log.org_database=     *db;
+  backup_log_ddl(&ddl_log);
+
 not_silent:
   if (!silent)
   {
@@ -740,6 +908,14 @@ mysql_alter_db_internal(THD *thd, const LEX_CSTRING *db,
     thd->variables.collation_database= thd->db_charset;
   }
 
+  /* Log command to ddl log */
+  backup_log_info ddl_log;
+  bzero(&ddl_log, sizeof(ddl_log));
+  ddl_log.query=                   { C_STRING_WITH_LEN("ALTER") };
+  ddl_log.org_storage_engine_name= { C_STRING_WITH_LEN("DATABASE") };
+  ddl_log.org_database=     *db;
+  backup_log_ddl(&ddl_log);
+
   if (mysql_bin_log.is_open())
   {
     int errcode= query_error_code(thd, TRUE); 
@@ -794,6 +970,56 @@ bool mysql_alter_db(THD *thd, const LEX_CSTRING *db,
 }
 
 
+/**
+   Drop database objects
+
+   @param thd              THD object
+   @param path             Path to database (for ha_drop_database)
+   @param db               Normalized database name
+   @param rm_mysql_schema  If the schema is 'mysql', in which case we don't
+                           log the query to binary log or delete related
+                           routines or events.
+*/
+
+void drop_database_objects(THD *thd, const LEX_CSTRING *path,
+                           const LEX_CSTRING *db,
+                           bool rm_mysql_schema)
+{
+  debug_crash_here("ddl_log_drop_before_ha_drop_database");
+
+  ha_drop_database(path->str);
+
+  /*
+    We temporarily disable the binary log while dropping the objects
+    in the database. Since the DROP DATABASE statement is always
+    replicated as a statement, execution of it will drop all objects
+    in the database on the slave as well, so there is no need to
+    replicate the removal of the individual objects in the database
+    as well.
+
+    This is more of a safety precaution, since normally no objects
+    should be dropped while the database is being cleaned, but in
+    the event that a change in the code to remove other objects is
+    made, these drops should still not be logged.
+  */
+
+  debug_crash_here("ddl_log_drop_before_drop_db_routines");
+
+  query_cache_invalidate1(thd, db->str);
+
+  if (!rm_mysql_schema)
+  {
+    tmp_disable_binlog(thd);
+    (void) sp_drop_db_routines(thd, db->str); /* @todo Do not ignore errors */
+#ifdef HAVE_EVENT_SCHEDULER
+    Events::drop_schema_events(thd, db->str);
+#endif
+    reenable_binlog(thd);
+  }
+  debug_crash_here("ddl_log_drop_after_drop_db_routines");
+}
+
+
 /**
   Drop all tables, routines and events in a database and the database itself.
 
@@ -810,41 +1036,31 @@ bool mysql_alter_db(THD *thd, const LEX_CSTRING *db,
 */
 
 static bool
-mysql_rm_db_internal(THD *thd, const LEX_CSTRING *db, bool if_exists, bool silent)
+mysql_rm_db_internal(THD *thd, const LEX_CSTRING *db, bool if_exists,
+                     bool silent)
 {
   ulong deleted_tables= 0;
   bool error= true, rm_mysql_schema;
   char	path[FN_REFLEN + 16];
   MY_DIR *dirp;
-  uint length;
+  uint path_length;
   TABLE_LIST *tables= NULL;
   TABLE_LIST *table;
+  DDL_LOG_STATE ddl_log_state;
   Drop_table_error_handler err_handler;
+  LEX_CSTRING rm_db;
+  char db_tmp[SAFE_NAME_LEN+1];
+  const char *dbnorm;
   DBUG_ENTER("mysql_rm_db");
 
-  char db_tmp[SAFE_NAME_LEN+1];
-  const char *dbnorm= normalize_db_name(db->str, db_tmp, sizeof(db_tmp));
+  dbnorm= normalize_db_name(db->str, db_tmp, sizeof(db_tmp));
+  lex_string_set(&rm_db, dbnorm);
+  bzero(&ddl_log_state, sizeof(ddl_log_state));
 
   if (lock_schema_name(thd, dbnorm))
     DBUG_RETURN(true);
 
-  length= build_table_filename(path, sizeof(path) - 1, db->str, "", "", 0);
-  strmov(path+length, MY_DB_OPT_FILE);		// Append db option file name
-  del_dbopt(path);				// Remove dboption hash entry
-  /*
-     Now remove the db.opt file.
-     The 'find_db_tables_and_rm_known_files' doesn't remove this file
-     if there exists a table with the name 'db', so let's just do it
-     separately. We know this file exists and needs to be deleted anyway.
-  */
-  if (mysql_file_delete_with_symlink(key_file_misc, path, "", MYF(0)) &&
-      my_errno != ENOENT)
-  {
-    my_error(EE_DELETE, MYF(0), path, my_errno);
-    DBUG_RETURN(true);
-  }
-    
-  path[length]= '\0';				// Remove file name
+  path_length= build_table_filename(path, sizeof(path) - 1, db->str, "", "", 0);
 
   /* See if the directory exists */
   if (!(dirp= my_dir(path,MYF(MY_DONT_SORT))))
@@ -895,7 +1111,10 @@ mysql_rm_db_internal(THD *thd, const LEX_CSTRING *db, bool if_exists, bool silen
     }
   }
 
-  /* mysql_ha_rm_tables() requires a non-null TABLE_LIST. */
+  /*
+    Close active HANDLER's for tables in the database.
+    Note that mysql_ha_rm_tables() requires a non-null TABLE_LIST.
+  */
   if (tables)
     mysql_ha_rm_tables(thd, tables);
 
@@ -905,44 +1124,59 @@ mysql_rm_db_internal(THD *thd, const LEX_CSTRING *db, bool if_exists, bool silen
   thd->push_internal_handler(&err_handler);
   if (!thd->killed &&
       !(tables &&
-        mysql_rm_table_no_locks(thd, tables, true, false, true, false, true,
-                                false)))
+        mysql_rm_table_no_locks(thd, tables, &rm_db, &ddl_log_state, true, false,
+                                true, false, true, false)))
   {
+    debug_crash_here("ddl_log_drop_after_drop_tables");
+
+    LEX_CSTRING cpath{ path, path_length};
+    ddl_log_drop_db(thd, &ddl_log_state, &rm_db, &cpath);
+
+    drop_database_objects(thd, &cpath, &rm_db, rm_mysql_schema);
+
     /*
-      We temporarily disable the binary log while dropping the objects
-      in the database. Since the DROP DATABASE statement is always
-      replicated as a statement, execution of it will drop all objects
-      in the database on the slave as well, so there is no need to
-      replicate the removal of the individual objects in the database
-      as well.
-
-      This is more of a safety precaution, since normally no objects
-      should be dropped while the database is being cleaned, but in
-      the event that a change in the code to remove other objects is
-      made, these drops should still not be logged.
+      Now remove the db.opt file.
+      The 'find_db_tables_and_rm_known_files' doesn't remove this file
+      if there exists a table with the name 'db', so let's just do it
+      separately. We know this file exists and needs to be deleted anyway.
     */
-
-    ha_drop_database(path);
-    tmp_disable_binlog(thd);
-    query_cache_invalidate1(thd, dbnorm);
-    if (!rm_mysql_schema)
+    debug_crash_here("ddl_log_drop_before_drop_option_file");
+    strmov(path+path_length, MY_DB_OPT_FILE);   // Append db option file name
+    if (mysql_file_delete_with_symlink(key_file_misc, path, "", MYF(0)) &&
+        my_errno != ENOENT)
     {
-      (void) sp_drop_db_routines(thd, dbnorm); /* @todo Do not ignore errors */
-#ifdef HAVE_EVENT_SCHEDULER
-      Events::drop_schema_events(thd, dbnorm);
-#endif
+      thd->pop_internal_handler();
+      my_error(EE_DELETE, MYF(0), path, my_errno);
+      error= true;
+      ddl_log_complete(&ddl_log_state);
+      goto end;
     }
-    reenable_binlog(thd);
+    del_dbopt(path);				// Remove dboption hash entry
+    path[path_length]= '\0';				// Remove file name
 
     /*
       If the directory is a symbolic link, remove the link first, then
       remove the directory the symbolic link pointed at
     */
+    debug_crash_here("ddl_log_drop_before_drop_dir");
     error= rm_dir_w_symlink(path, true);
+    debug_crash_here("ddl_log_drop_after_drop_dir");
   }
+
   thd->pop_internal_handler();
 
 update_binlog:
+  if (likely(!error))
+  {
+    /* Log command to ddl log */
+    backup_log_info ddl_log;
+    bzero(&ddl_log, sizeof(ddl_log));
+    ddl_log.query=                   { C_STRING_WITH_LEN("DROP") };
+    ddl_log.org_storage_engine_name= { C_STRING_WITH_LEN("DATABASE") };
+    ddl_log.org_database=     *db;
+    backup_log_ddl(&ddl_log);
+  }
+
   if (!silent && likely(!error))
   {
     const char *query;
@@ -955,6 +1189,7 @@ update_binlog:
     if (mysql_bin_log.is_open())
     {
       int errcode= query_error_code(thd, TRUE);
+      int res;
       Query_log_event qinfo(thd, query, query_length, FALSE, TRUE,
 			    /* suppress_use */ TRUE, errcode);
       /*
@@ -969,7 +1204,14 @@ update_binlog:
         These DDL methods and logging are protected with the exclusive
         metadata lock on the schema.
       */
-      if (mysql_bin_log.write(&qinfo))
+      debug_crash_here("ddl_log_drop_before_binlog");
+      thd->binlog_xid= thd->query_id;
+      ddl_log_update_xid(&ddl_log_state, thd->binlog_xid);
+      res= mysql_bin_log.write(&qinfo);
+      thd->binlog_xid= 0;
+      debug_crash_here("ddl_log_drop_after_binlog");
+
+      if (res)
       {
         error= true;
         goto exit;
@@ -1019,13 +1261,21 @@ update_binlog:
       *query_pos++ = ',';
     }
 
-    if (query_pos != query_data_start)
+    if (query_pos != query_data_start)          // If database was not empty
     {
+      int res;
       /*
         These DDL methods and logging are protected with the exclusive
         metadata lock on the schema.
       */
-      if (write_to_binlog(thd, query, (uint)(query_pos -1 - query), db->str, db->length))
+      debug_crash_here("ddl_log_drop_before_binlog");
+      thd->binlog_xid= thd->query_id;
+      ddl_log_update_xid(&ddl_log_state, thd->binlog_xid);
+      res= write_to_binlog(thd, query, (uint)(query_pos -1 - query), db->str,
+                           db->length);
+      thd->binlog_xid= 0;
+      debug_crash_here("ddl_log_drop_after_binlog");
+      if (res)
       {
         error= true;
         goto exit;
@@ -1034,6 +1284,7 @@ update_binlog:
   }
 
 exit:
+  ddl_log_complete(&ddl_log_state);
   /*
     If this database was the client's selected database, we silently
     change the client's selected database to nothing (to have an empty
@@ -1042,9 +1293,10 @@ exit:
   */
   if (unlikely(thd->db.str && cmp_db_names(&thd->db, db) && !error))
   {
-    mysql_change_db_impl(thd, NULL, 0, thd->variables.collation_server);
-    SESSION_TRACKER_CHANGED(thd, CURRENT_SCHEMA_TRACKER, NULL);
+    mysql_change_db_impl(thd, NULL, NO_ACL, thd->variables.collation_server);
+    thd->session_tracker.current_schema.mark_as_changed(thd);
   }
+end:
   my_dirend(dirp);
   DBUG_RETURN(error);
 }
@@ -1071,7 +1323,7 @@ static bool find_db_tables_and_rm_known_files(THD *thd, MY_DIR *dirp,
   DBUG_PRINT("enter",("path: %s", path));
 
   /* first, get the list of tables */
-  Dynamic_array<LEX_CSTRING*> files(dirp->number_of_files);
+  Dynamic_array<LEX_CSTRING*> files(PSI_INSTRUMENT_MEM, dirp->number_of_files);
   Discovered_table_list tl(thd, &files);
   if (ha_discover_table_names(thd, &db, dirp, &tl, true))
     DBUG_RETURN(1);
@@ -1102,9 +1354,9 @@ static bool find_db_tables_and_rm_known_files(THD *thd, MY_DIR *dirp,
                                                    (char*) table_list->table_name.str);
 
     table_list->alias= table_list->table_name;	// If lower_case_table_names=2
-    table_list->mdl_request.init(MDL_key::TABLE, table_list->db.str,
-                                 table_list->table_name.str, MDL_EXCLUSIVE,
-                                 MDL_TRANSACTION);
+    MDL_REQUEST_INIT(&table_list->mdl_request, MDL_key::TABLE,
+                     table_list->db.str, table_list->table_name.str,
+                     MDL_EXCLUSIVE, MDL_TRANSACTION);
     /* Link into list */
     (*tot_list_next_local)= table_list;
     (*tot_list_next_global)= table_list;
@@ -1175,22 +1427,24 @@ static bool find_db_tables_and_rm_known_files(THD *thd, MY_DIR *dirp,
     1 ERROR
 */
 
-static my_bool rm_dir_w_symlink(const char *org_path, my_bool send_error)
+my_bool rm_dir_w_symlink(const char *org_path, my_bool send_error)
 {
   char tmp_path[FN_REFLEN], *pos;
   char *path= tmp_path;
   DBUG_ENTER("rm_dir_w_symlink");
   unpack_filename(tmp_path, org_path);
-#ifdef HAVE_READLINK
-  int error;
-  char tmp2_path[FN_REFLEN];
 
-  /* Remove end FN_LIBCHAR as this causes problem on Linux in readlink */
+  /* Remove end FN_LIBCHAR as this causes problem on Linux and OS/2 */
   pos= strend(path);
   if (pos > path && pos[-1] == FN_LIBCHAR)
     *--pos=0;
 
-  if (unlikely((error= my_readlink(tmp2_path, path, MYF(MY_WME))) < 0))
+#ifdef HAVE_READLINK
+  int error;
+  char tmp2_path[FN_REFLEN];
+
+  if (unlikely((error= my_readlink(tmp2_path, path,
+                                   MYF(send_error ? MY_WME : 0))) < 0))
     DBUG_RETURN(1);
   if (likely(!error))
   {
@@ -1202,12 +1456,8 @@ static my_bool rm_dir_w_symlink(const char *org_path, my_bool send_error)
     path= tmp2_path;
   }
 #endif
-  /* Remove last FN_LIBCHAR to not cause a problem on OS/2 */
-  pos= strend(path);
 
-  if (pos > path && pos[-1] == FN_LIBCHAR)
-    *--pos=0;
-  if (unlikely(rmdir(path) < 0 && send_error))
+  if (unlikely(my_rmdir(path) < 0 && send_error))
   {
     my_error(ER_DB_DROP_RMDIR, MYF(0), path, errno);
     DBUG_RETURN(1);
@@ -1306,7 +1556,7 @@ err:
 
 static void mysql_change_db_impl(THD *thd,
                                  LEX_CSTRING *new_db_name,
-                                 ulong new_db_access,
+                                 privilege_t new_db_access,
                                  CHARSET_INFO *new_db_charset)
 {
   /* 1. Change current database in THD. */
@@ -1455,7 +1705,7 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
   LEX_CSTRING new_db_file_name;
 
   Security_context *sctx= thd->security_ctx;
-  ulong db_access= sctx->db_access;
+  privilege_t db_access(sctx->db_access);
   CHARSET_INFO *db_default_cl;
   DBUG_ENTER("mysql_change_db");
 
@@ -1472,7 +1722,7 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
         new_db_name->length == 0.
       */
 
-      mysql_change_db_impl(thd, NULL, 0, thd->variables.collation_server);
+      mysql_change_db_impl(thd, NULL, NO_ACL, thd->variables.collation_server);
 
       goto done;
     }
@@ -1501,8 +1751,8 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
     TODO: fix check_db_name().
   */
 
-  new_db_file_name.str= my_strndup(new_db_name->str, new_db_name->length,
-                                   MYF(MY_WME));
+  new_db_file_name.str= my_strndup(key_memory_THD_db, new_db_name->str,
+                                   new_db_name->length, MYF(MY_WME));
   new_db_file_name.length= new_db_name->length;
 
   if (new_db_file_name.str == NULL)
@@ -1524,7 +1774,7 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
     my_free(const_cast<char*>(new_db_file_name.str));
 
     if (force_switch)
-      mysql_change_db_impl(thd, NULL, 0, thd->variables.collation_server);
+      mysql_change_db_impl(thd, NULL, NO_ACL, thd->variables.collation_server);
 
     DBUG_RETURN(ER_WRONG_DB_NAME);
   }
@@ -1576,7 +1826,7 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
 
       /* Change db to NULL. */
 
-      mysql_change_db_impl(thd, NULL, 0, thd->variables.collation_server);
+      mysql_change_db_impl(thd, NULL, NO_ACL, thd->variables.collation_server);
 
       /* The operation succeed. */
       goto done;
@@ -1604,8 +1854,8 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
   mysql_change_db_impl(thd, &new_db_file_name, db_access, db_default_cl);
 
 done:
-  SESSION_TRACKER_CHANGED(thd, CURRENT_SCHEMA_TRACKER, NULL);
-  SESSION_TRACKER_CHANGED(thd, SESSION_STATE_CHANGE_TRACKER, NULL);
+  thd->session_tracker.current_schema.mark_as_changed(thd);
+  thd->session_tracker.state_change.mark_as_changed(thd);
   DBUG_RETURN(0);
 }
 
@@ -1764,7 +2014,7 @@ bool mysql_upgrade_db(THD *thd, const LEX_CSTRING *old_db)
   }
 
   if ((table_list= thd->lex->query_tables) &&
-      (error= mysql_rename_tables(thd, table_list, 1)))
+      (error= mysql_rename_tables(thd, table_list, 1, 0)))
   {
     /*
       Failed to move all tables from the old database to the new one.
@@ -1783,7 +2033,7 @@ bool mysql_upgrade_db(THD *thd, const LEX_CSTRING *old_db)
     length= build_table_filename(path, sizeof(path)-1, new_db.str, "", "", 0);
     if (length && path[length-1] == FN_LIBCHAR)
       path[length-1]=0;                            // remove ending '\'
-    rmdir(path);
+    my_rmdir(path);
     goto exit;
   }
 
@@ -1878,20 +2128,34 @@ exit:
     TRUE    The directory does not exist.
 */
 
+
 bool check_db_dir_existence(const char *db_name)
 {
   char db_dir_path[FN_REFLEN + 1];
   uint db_dir_path_len;
 
+  if (dbname_cache->contains(db_name))
+    return 0;
+
   db_dir_path_len= build_table_filename(db_dir_path, sizeof(db_dir_path) - 1,
                                         db_name, "", "", 0);
 
   if (db_dir_path_len && db_dir_path[db_dir_path_len - 1] == FN_LIBCHAR)
     db_dir_path[db_dir_path_len - 1]= 0;
 
-  /* Check access. */
+  /*
+    Check access.
 
-  return my_access(db_dir_path, F_OK);
+    The locking is to prevent creating permanent stale
+    entries for deleted databases, in case of
+    race condition with my_rmdir.
+  */
+  mysql_rwlock_rdlock(&rmdir_lock);
+  int ret= my_access(db_dir_path, F_OK);
+  if (!ret)
+    dbname_cache->insert(db_name);
+  mysql_rwlock_unlock(&rmdir_lock);
+  return ret;
 }
 
 
