@@ -655,14 +655,29 @@ static
 bool mysql_derived_prepare(THD *thd, LEX *lex, TABLE_LIST *derived)
 {
   SELECT_LEX_UNIT *unit= derived->get_unit();
-  bool res= FALSE;
+  SELECT_LEX *first_select;
+  bool res= FALSE, keep_row_order;
   DBUG_ENTER("mysql_derived_prepare");
   DBUG_PRINT("enter", ("unit: %p  table_list: %p  alias: '%s'",
                        unit, derived, derived->alias.str));
   if (!unit)
     DBUG_RETURN(FALSE);
 
-  SELECT_LEX *first_select= unit->first_select();
+  first_select= unit->first_select();
+  /*
+    If rownum() is used we have to preserve the insert row order
+    to make GROUP BY and ORDER BY with filesort work.
+
+    SELECT * from (SELECT a,b from t1 ORDER BY a)) WHERE rownum <= 0;
+
+    When rownum is not used the optimizer will skip the ORDER BY clause.
+    With rownum we have to keep the ORDER BY as this is what is expected.
+    We also have to create any sort result temporary table in such a way
+    that the inserted row order is maintained.
+  */
+  keep_row_order= (thd->lex->with_rownum &&
+                   (first_select->group_list.elements ||
+                    first_select->order_list.elements));
 
   if (derived->is_recursive_with_table() &&
       !derived->is_with_table_recursive_reference() &&
@@ -699,7 +714,8 @@ bool mysql_derived_prepare(THD *thd, LEX *lex, TABLE_LIST *derived)
                                   (first_select->options |
                                    thd->variables.option_bits |
                                    TMP_TABLE_ALL_COLUMNS),
-                                  &derived->alias, FALSE, FALSE, FALSE, 0);
+                                  &derived->alias, FALSE, FALSE,
+                                  keep_row_order, 0);
     thd->create_tmp_table_for_derived= FALSE;
 
     if (likely(!res) && !derived->table)
@@ -855,7 +871,7 @@ bool mysql_derived_prepare(THD *thd, LEX *lex, TABLE_LIST *derived)
                                                    thd->variables.option_bits |
                                                    TMP_TABLE_ALL_COLUMNS),
                                                    &derived->alias,
-                                                   FALSE, FALSE, FALSE,
+                                                   FALSE, FALSE, keep_row_order,
                                                    0))
   { 
     thd->create_tmp_table_for_derived= FALSE;
@@ -1494,7 +1510,8 @@ bool pushdown_cond_for_derived(THD *thd, Item *cond, TABLE_LIST *derived)
     DBUG_RETURN(false);
 
   /* Do not push conditions into unit with global ORDER BY ... LIMIT */
-  if (unit->fake_select_lex && unit->fake_select_lex->explicit_limit)
+  if (unit->fake_select_lex &&
+      unit->fake_select_lex->limit_params.explicit_limit)
     DBUG_RETURN(false);
 
   /* Check whether any select of 'unit' allows condition pushdown */
