@@ -44,6 +44,8 @@
 
 #include "wsrep_mysqld.h"
 
+#include "scope.h"  // scope_exit
+
 extern "C" int _my_b_net_read(IO_CACHE *info, uchar *Buffer, size_t Count);
 
 class XML_TAG {
@@ -51,11 +53,11 @@ public:
   int level;
   String field;
   String value;
-  XML_TAG(int l, String f, String v);
+  XML_TAG(int l, const String &f, const String &v);
 };
 
 
-XML_TAG::XML_TAG(int l, String f, String v)
+XML_TAG::XML_TAG(int l, const String &f, const String &v)
 {
   level= l;
   field.append(f);
@@ -105,41 +107,23 @@ public:
 class Wsrep_load_data_split
 {
 public:
-  Wsrep_load_data_split(THD *thd, TABLE *table)
+  Wsrep_load_data_split(THD *thd)
     : m_thd(thd)
-    , m_load_data_splitting(false)
+    , m_load_data_splitting(wsrep_load_data_splitting)
     , m_fragment_unit(thd->wsrep_trx().streaming_context().fragment_unit())
     , m_fragment_size(thd->wsrep_trx().streaming_context().fragment_size())
   {
-    /*
-      We support load data splitting for InnoDB only as it will use
-      streaming replication (SR).
-    */
-    if (WSREP(thd) && wsrep_load_data_splitting)
+    if (WSREP(m_thd) && m_load_data_splitting)
     {
-      handlerton *ht= table->s->db_type();
-      // For partitioned tables find underlying hton
-      if (table->file->partition_ht())
-        ht= table->file->partition_ht();
-      if (ht->db_type != DB_TYPE_INNODB)
-      {
-        push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
-                            ER_NOT_SUPPORTED_YET,
-                            "wsrep_load_data_splitting for other than InnoDB tables");
-      }
-      else
-      {
-        /* Override streaming settings with backward compatible values for
-           load data splitting */
-        m_thd->wsrep_cs().streaming_params(wsrep::streaming_context::row, 10000);
-        m_load_data_splitting= true;
-      }
+      /* Override streaming settings with backward compatible values for
+         load data splitting */
+      m_thd->wsrep_cs().streaming_params(wsrep::streaming_context::row, 10000);
     }
   }
 
   ~Wsrep_load_data_split()
   {
-    if (m_load_data_splitting)
+    if (WSREP(m_thd) && m_load_data_splitting)
     {
       /* Restore original settings */
       m_thd->wsrep_cs().streaming_params(m_fragment_unit, m_fragment_size);
@@ -364,7 +348,6 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   bool is_concurrent;
 #endif
   const char *db= table_list->db.str;		// This is never null
-
   /*
     If path for file is not defined, we will use the current database.
     If this is not set, we will use the directory where the table to be
@@ -375,6 +358,9 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   bool transactional_table __attribute__((unused));
   DBUG_ENTER("mysql_load");
 
+#ifdef WITH_WSREP
+  Wsrep_load_data_split wsrep_load_data_split(thd);
+#endif /* WITH_WSREP */
   /*
     Bug #34283
     mysqlbinlog leaves tmpfile after termination if binlog contains
@@ -439,11 +425,6 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   {
     DBUG_RETURN(TRUE);
   }
-
-#ifdef WITH_WSREP
-  Wsrep_load_data_split wsrep_load_data_split(thd, table_list->table);
-#endif /* WITH_WSREP */
-
   thd_proc_info(thd, "Executing");
   /*
     Let us emit an error if we are loading data to table which is used
@@ -469,6 +450,12 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   if (check_duplic_insert_without_overlaps(thd, table, handle_duplicates) != 0)
     DBUG_RETURN(true);
 
+  auto scope_cleaner = make_scope_exit(
+    [&fields_vars]() {
+      fields_vars.empty();
+    }
+  );
+
   if (!fields_vars.elements)
   {
     Field_iterator_table_ref field_iterator;
@@ -496,6 +483,7 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
   }
   else
   {						// Part field list
+    scope_cleaner.release();
     /* TODO: use this conds for 'WITH CHECK OPTIONS' */
     if (setup_fields(thd, Ref_ptr_array(),
                      fields_vars, MARK_COLUMNS_WRITE, 0, NULL, 0) ||
@@ -592,7 +580,7 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
       DBUG_RETURN(TRUE);
     }
 
-#if !defined(__WIN__) && ! defined(__NETWARE__)
+#if !defined(_WIN32)
     MY_STAT stat_info;
     if (!my_stat(name, &stat_info, MYF(MY_WME)))
       DBUG_RETURN(TRUE);
@@ -686,6 +674,7 @@ int mysql_load(THD *thd, const sql_exchange *ex, TABLE_LIST *table_list,
     }
     table->file->prepare_for_insert(create_lookup_handler);
     thd_progress_init(thd, 2);
+    fix_rownum_pointers(thd, thd->lex->current_select, &info.copied);
     if (table_list->table->validate_default_values_of_unset_fields(thd))
     {
       read_info.error= true;
@@ -893,7 +882,7 @@ static bool write_execute_load_query_log_event(THD *thd, const sql_exchange* ex,
      */
     qualify_db= db_arg;
   }
-  lle.print_query(thd, FALSE, (const char *) ex->cs?ex->cs->csname:NULL,
+  lle.print_query(thd, FALSE, (const char*) ex->cs ? ex->cs->cs_name.str : NULL,
                   &query_str, &fname_start, &fname_end, qualify_db);
 
   /*
@@ -903,18 +892,18 @@ static bool write_execute_load_query_log_event(THD *thd, const sql_exchange* ex,
   {
     List_iterator<Item>  li(thd->lex->field_list);
 
-    query_str.append(" (");
+    query_str.append(STRING_WITH_LEN(" ("));
     n= 0;
 
     while ((item= li++))
     {
       if (n++)
-        query_str.append(", ");
+        query_str.append(STRING_WITH_LEN(", "));
       const Load_data_outvar *var= item->get_load_data_outvar();
       DBUG_ASSERT(var);
       var->load_data_print_for_log_event(thd, &query_str);
     }
-    query_str.append(")");
+    query_str.append(')');
   }
 
   if (!thd->lex->update_list.is_empty())
