@@ -20,7 +20,7 @@
 
 #include "ma_fulltext.h"
 #include <my_check_opt.h>
-#if defined(MSDOS) || defined(__WIN__)
+#if defined(_WIN32)
 #include <fcntl.h>
 #else
 #include <stddef.h>
@@ -177,8 +177,7 @@ int _ma_create_index_by_sort(MARIA_SORT_PARAM *info, my_bool no_messages,
 	maxbuffer_org= maxbuffer;
 	if (memavl < sizeof(BUFFPEK) * maxbuffer ||
 	    (keys= (memavl-sizeof(BUFFPEK)*maxbuffer)/
-             (sort_length+sizeof(char*))) <= 1 ||
-            keys < maxbuffer)
+             (sort_length+sizeof(char*))) <= 1)
 	{
 	  _ma_check_print_error(info->sort_info->param,
                                 "aria_sort_buffer_size is too small. Current aria_sort_buffer_size: %llu  rows: %llu  sort_length: %u",
@@ -187,17 +186,26 @@ int _ma_create_index_by_sort(MARIA_SORT_PARAM *info, my_bool no_messages,
           my_errno= ENOMEM;
 	  goto err;
 	}
+        if (keys < maxbuffer)
+        {
+          /*
+            There must be sufficient memory for at least one key per BUFFPEK,
+            otherwise repair by sort/parallel repair cannot operate.
+          */
+          maxbuffer= (uint) keys;
+          break;
+        }
       }
       while ((maxbuffer= (uint) (records/(keys-1)+1)) != maxbuffer_org);
     }
 
     if ((sort_keys= ((uchar**)
-                     my_malloc((size_t) (keys*(sort_length+sizeof(char*))+
+                     my_malloc(PSI_INSTRUMENT_ME, (size_t) (keys*(sort_length+sizeof(char*))+
                                          HA_FT_MAXBYTELEN),
                                MYF(0)))))
     {
-      if (my_init_dynamic_array(&buffpek, sizeof(BUFFPEK), maxbuffer,
-                                MY_MIN(maxbuffer/2, 1000), MYF(0)))
+      if (my_init_dynamic_array(PSI_INSTRUMENT_ME, &buffpek, sizeof(BUFFPEK),
+                                maxbuffer, MY_MIN(maxbuffer/2, 1000), MYF(0)))
       {
 	my_free(sort_keys);
         sort_keys= 0;
@@ -231,6 +239,20 @@ int _ma_create_index_by_sort(MARIA_SORT_PARAM *info, my_bool no_messages,
       == HA_POS_ERROR)
     goto err; /* purecov: tested */
 
+  if (maxbuffer >= keys)
+  {
+    /*
+      merge_many_buff will crash if maxbuffer > keys as then we cannot store in memory
+      the keys for each buffer.
+    */
+    keys= maxbuffer + 1;
+    if (!(sort_keys= ((uchar **)
+                      my_realloc(PSI_INSTRUMENT_ME, sort_keys,
+                                 (size_t) (keys*(sort_length+sizeof(char*))+
+                                           HA_FT_MAXBYTELEN), MYF(MY_FREE_ON_ERROR)))))
+      goto err;
+  }
+
   info->sort_info->param->stage++;                         /* Merge stage */
 
   if (maxbuffer == 0)
@@ -426,11 +448,11 @@ static my_bool _ma_thr_find_all_keys_exec(MARIA_SORT_PARAM* sort_param)
       while ((maxbuffer= (uint) (idx/(keys-1)+1)) != maxbuffer_org);
     }
     if ((sort_keys= (uchar **)
-         my_malloc((size_t)(keys*(sort_length+sizeof(char*))+
+         my_malloc(PSI_INSTRUMENT_ME, (size_t)(keys*(sort_length+sizeof(char*))+
                    ((sort_param->keyinfo->flag & HA_FULLTEXT) ?
                     HA_FT_MAXBYTELEN : 0)), MYF(0))))
     {
-      if (my_init_dynamic_array(&sort_param->buffpek, sizeof(BUFFPEK),
+      if (my_init_dynamic_array(PSI_INSTRUMENT_ME, &sort_param->buffpek, sizeof(BUFFPEK),
                              maxbuffer, MY_MIN(maxbuffer / 2, 1000), MYF(0)))
       {
         my_free(sort_keys);
@@ -626,7 +648,7 @@ int _ma_thr_write_keys(MARIA_SORT_PARAM *sort_param)
         length=(size_t)param->sort_buffer_length;
         while (length >= MARIA_MIN_SORT_MEMORY)
         {
-          if ((mergebuf= my_malloc((size_t) length, MYF(0))))
+          if ((mergebuf= my_malloc(PSI_INSTRUMENT_ME, (size_t) length, MYF(0))))
               break;
           length=length*3/4;
         }
