@@ -28,31 +28,56 @@
 #define SPIDER_DB_WRAPPER_MYSQL "mysql"
 #define SPIDER_DB_WRAPPER_MARIADB "mariadb"
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100204
 #define PLUGIN_VAR_CAN_MEMALLOC
 /*
+#define ITEM_FUNC_CASE_PARAMS_ARE_PUBLIC
 #define HASH_UPDATE_WITH_HASH_VALUE
 */
+#else
+#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+#define HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
+#endif
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100002
 #define SPIDER_HAS_DISCOVER_TABLE_STRUCTURE
 #define SPIDER_HAS_APPEND_FOR_SINGLE_QUOTE
 #define SPIDER_HAS_SHOW_SIMPLE_FUNC
 #define SPIDER_HAS_JT_HASH_INDEX_MERGE
 #define SPIDER_HAS_EXPR_CACHE_ITEM
+#else
+#define SPIDER_NEED_CHECK_CONDITION_AT_CHECKING_DIRECT_ORDER_LIMIT
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100007
 #define SPIDER_ITEM_HAS_CMP_TYPE
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100004
 #define SPIDER_HAS_TIME_STATUS
 #define SPIDER_HAS_DECIMAL_OPERATION_RESULTS_VALUE_TYPE
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100014
 #define SPIDER_ITEM_STRING_WITHOUT_SET_STR_WITH_COPY
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100100
 #define SPIDER_ITEM_STRING_WITHOUT_SET_STR_WITH_COPY_AND_THDPTR
+#endif
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100108
 #define SPIDER_HAS_GROUP_BY_HANDLER
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100200
 #define SPIDER_ORDER_HAS_ENUM_ORDER
+#endif
 
+#if defined(MARIADB_BASE_VERSION)
 #define SPIDER_ITEM_GEOFUNC_NAME_HAS_MBR
 #define SPIDER_HANDLER_AUTO_REPAIR_HAS_ERROR
+#endif
 
 class spider_db_conn;
 typedef spider_db_conn SPIDER_DB_CONN;
@@ -114,6 +139,7 @@ typedef st_spider_result SPIDER_RESULT;
 #define SPIDER_SQL_HS_LTEQUAL_STR "<="
 #define SPIDER_SQL_HS_LTEQUAL_LEN (sizeof(SPIDER_SQL_HS_LTEQUAL_STR) - 1)
 
+#ifdef ITEM_FUNC_CASE_PARAMS_ARE_PUBLIC
 #define SPIDER_SQL_CASE_STR "case "
 #define SPIDER_SQL_CASE_LEN (sizeof(SPIDER_SQL_CASE_STR) - 1)
 #define SPIDER_SQL_WHEN_STR " when "
@@ -124,6 +150,7 @@ typedef st_spider_result SPIDER_RESULT;
 #define SPIDER_SQL_ELSE_LEN (sizeof(SPIDER_SQL_ELSE_STR) - 1)
 #define SPIDER_SQL_END_STR " end"
 #define SPIDER_SQL_END_LEN (sizeof(SPIDER_SQL_END_STR) - 1)
+#endif
 
 #define SPIDER_SQL_USING_STR " using "
 #define SPIDER_SQL_USING_LEN (sizeof(SPIDER_SQL_USING_STR) - 1)
@@ -327,11 +354,6 @@ public:
     const char *str,
     uint32 arg_length
   );
-  void set_quick(
-    char *str,
-    uint32 arg_length,
-    CHARSET_INFO *cs
-  );
   bool set_int(
     longlong num,
     bool unsigned_flag,
@@ -897,6 +919,9 @@ public:
   virtual bool tables_on_different_db_are_joinable();
   virtual bool socket_has_default_value();
   virtual bool database_has_default_value();
+  virtual bool default_file_has_default_value();
+  virtual bool host_has_default_value();
+  virtual bool port_has_default_value();
   virtual bool append_charset_name_before_string();
   virtual uint limit_mode();
 };
@@ -1143,6 +1168,10 @@ public:
     Time_zone *time_zone,
     int *need_mon
   ) = 0;
+  virtual bool set_loop_check_in_bulk_sql();
+  virtual int set_loop_check(
+    int *need_mon
+  );
   virtual int fin_loop_check();
   virtual int show_master_status(
     SPIDER_TRX *trx,
