@@ -191,6 +191,12 @@ public:
     spider_string *str
   ) override;
 #endif
+  bool tables_on_different_db_are_joinable() override;
+  bool socket_has_default_value() override;
+  bool database_has_default_value() override;
+  bool default_file_has_default_value() override;
+  bool host_has_default_value() override;
+  bool port_has_default_value() override;
   bool append_charset_name_before_string() override;
 };
 
@@ -527,6 +533,11 @@ public:
     Time_zone *time_zone,
     int *need_mon
   );
+  bool set_loop_check_in_bulk_sql();
+  int set_loop_check(
+    int *need_mon
+  );
+  int fin_loop_check();
   int exec_simple_sql_with_result(
     SPIDER_TRX *trx,
     SPIDER_SHARE *share,
