@@ -55,6 +55,8 @@ static LEX_STRING spider_init_queries[] = {
     "  default_file text,"
     "  default_group char(64) default null,"
     "  dsn char(64) default null,"
+    "  filedsn text default null,"
+    "  driver char(64) default null,"
     "  key idx1 (data, format_id, gtrid_length, host)"
     ") engine=MyISAM default charset=utf8 collate=utf8_bin"
   )},
@@ -79,6 +81,8 @@ static LEX_STRING spider_init_queries[] = {
     "  default_file text,"
     "  default_group char(64) default null,"
     "  dsn char(64) default null,"
+    "  filedsn text default null,"
+    "  driver char(64) default null,"
     "  thread_id int default null,"
     "  status char(8) not null default '',"
     "  failed_time timestamp not null default current_timestamp,"
@@ -108,6 +112,8 @@ static LEX_STRING spider_init_queries[] = {
     "  default_file text,"
     "  default_group char(64) default null,"
     "  dsn char(64) default null,"
+    "  filedsn text default null,"
+    "  driver char(64) default null,"
     "  tgt_db_name char(64) default null,"
     "  tgt_table_name char(64) default null,"
     "  link_status tinyint not null default 1,"
@@ -140,6 +146,8 @@ static LEX_STRING spider_init_queries[] = {
     "  default_file text,"
     "  default_group char(64) default null,"
     "  dsn char(64) default null,"
+    "  filedsn text default null,"
+    "  driver char(64) default null,"
     "  primary key (db_name, table_name, link_id, sid)"
     ") engine=MyISAM default charset=utf8 collate=utf8_bin"
   )},
@@ -644,6 +652,46 @@ static LEX_STRING spider_init_queries[] = {
     "alter table mysql.spider_xa_member"
     "  add column if not exists dsn char(64) default null after default_group;"
   )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_link_mon_servers"
+    "  add column if not exists filedsn text default null after dsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_tables"
+    "  add column if not exists filedsn text default null after dsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_xa_failed_log"
+    "  add column if not exists filedsn text default null after dsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_xa_member"
+    "  add column if not exists filedsn text default null after dsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_link_mon_servers"
+    "  add column if not exists driver char(64) default null after filedsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_tables"
+    "  add column if not exists driver char(64) default null after filedsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_xa_failed_log"
+    "  add column if not exists driver char(64) default null after filedsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_xa_member"
+    "  add column if not exists driver char(64) default null after filedsn,"
+    "  algorithm=copy, lock=shared;"
+  )},
   {C_STRING_WITH_LEN(
     "set @win_plugin := IF(@@version_compile_os like 'Win%', 1, 0);"
   )},
