@@ -203,7 +203,7 @@ void *spider_alloc_mem(
   uchar *ptr;
   DBUG_ENTER("spider_alloc_mem");
   size += ALIGN_SIZE(sizeof(uint)) + ALIGN_SIZE(sizeof(uint));
-  if (!(ptr = (uchar *) my_malloc(size, my_flags)))
+  if (!(ptr = (uchar *) my_malloc(PSI_INSTRUMENT_ME, size, my_flags)))
     DBUG_RETURN(NULL);
 
   spider_alloc_mem_calc(trx, id, func_name, file_name, line_no, size);
@@ -233,7 +233,7 @@ void *spider_bulk_alloc_mem(
     total_size += ALIGN_SIZE(va_arg(args, uint));
   va_end(args);
 
-  if (!(top_ptr = (uchar *) my_malloc(total_size, my_flags)))
+  if (!(top_ptr = (uchar *) my_malloc(PSI_INSTRUMENT_ME, total_size, my_flags)))
     DBUG_RETURN(NULL);
 
   spider_alloc_mem_calc(trx, id, func_name, file_name, line_no, total_size);
@@ -292,7 +292,7 @@ spider_string::spider_string(
 spider_string::spider_string(
   const char *str,
   CHARSET_INFO *cs
-) : str(str, cs), next(NULL)
+) : str(str, strlen(str), cs), next(NULL)
 {
   DBUG_ENTER("spider_string::spider_string");
   DBUG_PRINT("info",("spider this=%p", this));
@@ -476,9 +476,10 @@ char *spider_string::c_ptr_safe()
 
 LEX_STRING spider_string::lex_string() const
 {
+  LEX_STRING res= { (char*) str.ptr(), str.length() };
   DBUG_ENTER("spider_string::lex_string");
   DBUG_PRINT("info",("spider this=%p", this));
-  DBUG_RETURN(str.lex_string());
+  DBUG_RETURN(res);
 }
 
 void spider_string::set(
@@ -540,21 +541,6 @@ bool spider_string::set_ascii(
   DBUG_RETURN(res);
 }
 
-void spider_string::set_quick(
-  char *str,
-  uint32 arg_length,
-  CHARSET_INFO *cs
-) {
-  DBUG_ENTER("spider_string::set_quick");
-  DBUG_PRINT("info",("spider this=%p", this));
-  DBUG_ASSERT(mem_calc_inited);
-  DBUG_ASSERT((!current_alloc_mem && !this->str.is_alloced()) ||
-    current_alloc_mem == this->str.alloced_length());
-  this->str.set_quick(str, arg_length, cs);
-  SPIDER_STRING_CALC_MEM;
-  DBUG_VOID_RETURN;
-}
-
 bool spider_string::set_int(
   longlong num,
   bool unsigned_flag,
@@ -867,7 +853,7 @@ bool spider_string::append(
   DBUG_ASSERT(mem_calc_inited);
   DBUG_ASSERT((!current_alloc_mem && !str.is_alloced()) ||
     current_alloc_mem == str.alloced_length());
-  bool res = str.append(s);
+  bool res = str.append(s, strlen(s));
   SPIDER_STRING_CALC_MEM;
   DBUG_RETURN(res);
 }
@@ -1213,13 +1199,15 @@ void spider_string::append_escape_string(
   const char *st,
   uint len
 ) {
+  my_bool overflow;
   DBUG_ENTER("spider_string::append_escape_string");
   DBUG_PRINT("info",("spider this=%p", this));
   DBUG_ASSERT(mem_calc_inited);
   DBUG_ASSERT((!current_alloc_mem && !str.is_alloced()) ||
     current_alloc_mem == str.alloced_length());
   str.length(str.length() + escape_string_for_mysql(
-    str.charset(), (char *) str.ptr() + str.length(), 0, st, len));
+    str.charset(), (char *) str.ptr() + str.length(), 0, st, len,
+                                                    &overflow));
   DBUG_VOID_RETURN;
 }
 
@@ -1228,13 +1216,14 @@ void spider_string::append_escape_string(
   uint len,
   CHARSET_INFO *cs
 ) {
+  my_bool overflow;
   DBUG_ENTER("spider_string::append_escape_string");
   DBUG_PRINT("info",("spider this=%p", this));
   DBUG_ASSERT(mem_calc_inited);
   DBUG_ASSERT((!current_alloc_mem && !str.is_alloced()) ||
     current_alloc_mem == str.alloced_length());
   str.length(str.length() + escape_string_for_mysql(
-    cs, (char *) str.ptr() + str.length(), 0, st, len));
+    cs, (char *) str.ptr() + str.length(), 0, st, len, &overflow));
   DBUG_VOID_RETURN;
 }
 
@@ -1283,7 +1272,7 @@ bool spider_string::append_for_single_quote(
   DBUG_ASSERT((!current_alloc_mem && !str.is_alloced()) ||
     current_alloc_mem == str.alloced_length());
 #ifdef SPIDER_HAS_APPEND_FOR_SINGLE_QUOTE
-  bool res = str.append_for_single_quote(st);
+  bool res = str.append_for_single_quote(st, strlen(st));
 #else
   String ststr(st, str.charset());
   bool res = append_escaped(&str, &ststr);
