@@ -1,4 +1,5 @@
-/* Copyright (C) 2009-2018 Kentoku Shiba
+/* Copyright (C) 2009-2019 Kentoku Shiba
+   Copyright (C) 2019 MariaDB corp
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -17,11 +18,16 @@
 #include <my_global.h>
 #include "mysql_version.h"
 #include "spd_environ.h"
+#if MYSQL_VERSION_ID < 50500
+#include "mysql_priv.h"
+#include <mysql/plugin.h>
+#else
 #include "sql_priv.h"
 #include "probes_mysql.h"
 #include "sql_class.h"
 #include "sql_partition.h"
 #include "sql_acl.h"
+#endif
 #include "spd_err.h"
 #include "spd_param.h"
 #include "spd_db_include.h"
@@ -49,8 +55,6 @@ extern PSI_mutex_key spd_key_mutex_mon_list_update_status;
 extern PSI_mutex_key spd_key_mutex_mon_table_cache;
 #endif
 
-/* Array (of size `spider_udf_table_mon_mutex_count') of hashes of
-`SPIDER_TABLE_MON_LIST'. */
 HASH *spider_udf_table_mon_list_hash;
 uint spider_udf_table_mon_list_hash_id;
 const char *spider_udf_table_mon_list_hash_func_name;
@@ -60,43 +64,23 @@ pthread_mutex_t *spider_udf_table_mon_mutexes;
 pthread_cond_t *spider_udf_table_mon_conds;
 
 pthread_mutex_t spider_mon_table_cache_mutex;
-/* A cache to store distinct SPIDER_MON_KEYs with db name, table name
-and link id read from mysql.spider_link_mon_servers table. Initialised
-and populated in spider_init_ping_table_mon_cache(), and used in
-spider_ping_table_cache_compare(). The udf
-spider_flush_table_mon_cache is used to flag a initialisation. */
 DYNAMIC_ARRAY spider_mon_table_cache;
 uint spider_mon_table_cache_id;
 const char *spider_mon_table_cache_func_name;
 const char *spider_mon_table_cache_file_name;
 ulong spider_mon_table_cache_line_no;
-/* The mon table cache version, initialised at 0, and always no
-greater than spider_mon_table_cache_version_req. When the inequality
-is strict, an initialisation of spider_mon_table_cache will be
-triggered. */
-volatile ulonglong spider_mon_table_cache_version;
-/* The required mon table cache version, incremented by one by the
-udf spider_flush_table_mon_cache */
-volatile ulonglong spider_mon_table_cache_version_req;
+volatile ulonglong spider_mon_table_cache_version = 0;
+volatile ulonglong spider_mon_table_cache_version_req = 1;
 
- /* Get or create a `SPIDER_TABLE_MON_LIST' for a key `str' */
 SPIDER_TABLE_MON_LIST *spider_get_ping_table_mon_list(
   SPIDER_TRX *trx,
   THD *thd,
-  spider_string *str,           /* The key to search in
-                                  `spider_udf_table_mon_list_hash',
-                                  usually in the format of
-                                  "./$db_name/$table_name000000000$link_idx" */
+  spider_string *str,
   uint conv_name_length,
   int link_idx,
   char *static_link_id,
   uint static_link_id_length,
-  uint32 server_id,             /* The server id of the monitor
-                                  server, used for creating a new
-                                  table mon list having a
-                                  `SPIDER_TABLE_MON' corresponding to
-                                  the server id as the `current'
-                                  field */
+  uint32 server_id,
   bool need_lock,
   int *error_num
 ) {
@@ -108,7 +92,6 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_mon_list(
   my_hash_value_type hash_value;
 #endif
   DBUG_ENTER("spider_get_ping_table_mon_list");
-  /* Reset the cache if the version does not match the requirement */
   if (spider_mon_table_cache_version != spider_mon_table_cache_version_req)
   {
     SPD_INIT_ALLOC_ROOT(&mem_root, 4096, 0, MYF(MY_WME));
@@ -121,9 +104,6 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_mon_list(
     free_root(&mem_root, MYF(0));
   }
 
-  /* Search for the table mon list in the hash, if one is not found or
-  if it is found but has the wrong cache version, create and
-  initialise a new one. */
   mutex_hash = spider_udf_calc_hash(str->c_ptr(),
     spider_param_udf_table_mon_mutex_count());
   DBUG_PRINT("info",("spider hash key=%s", str->c_ptr()));
@@ -150,15 +130,12 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_mon_list(
   )
 #endif
   {
-    /* If table_mon_list is found but the cache version does not
-    match, remove it from the hash and free it. */
     if (
       table_mon_list &&
       table_mon_list->mon_table_cache_version != mon_table_cache_version
     )
       spider_release_ping_table_mon_list_loop(mutex_hash, table_mon_list);
-    /* create and initialise `table_mon_list' and insert it into the
-    hash */
+
     if (!(table_mon_list = spider_get_ping_table_tgt(thd, str->c_ptr(),
       conv_name_length, link_idx, static_link_id, static_link_id_length,
       server_id, str, need_lock, error_num)))
@@ -301,14 +278,6 @@ int spider_release_ping_table_mon_list(
   DBUG_RETURN(0);
 }
 
-/*
-  Look for a `SPIDER_MON_KEY` in `spider_mon_table_cache' whose db and
-  table name and link_idx matching `name' and `link_idx' with wild
-  card matching. If a match is found, create `SPIDER_TABLE_MON's from
-  all rows in mysql.spider_link_mon_servers that match the info in the
-  `SPIDER_MON_KEY' and populate the `table_mon_list' with these
-  `SPIDER_TABLE_MON's.
-*/
 int spider_get_ping_table_mon(
   THD *thd,
   SPIDER_TABLE_MON_LIST *table_mon_list,
@@ -321,7 +290,7 @@ int spider_get_ping_table_mon(
 ) {
   int error_num;
   TABLE *table_link_mon = NULL;
-  Open_tables_backup open_tables_backup;
+  SPIDER_Open_tables_backup open_tables_backup;
   char table_key[MAX_KEY_LENGTH];
   SPIDER_TABLE_MON *table_mon, *table_mon_prev = NULL;
   SPIDER_SHARE *tmp_share;
@@ -385,8 +354,6 @@ int spider_get_ping_table_mon(
   goto error;
 
 create_table_mon:
-  /* Find the first row in mysql.spider_link_mon_servers matching the
-  db name, table name and link_idx */
   if ((error_num = spider_get_sys_table_by_idx(table_link_mon, table_key,
     table_link_mon->s->primary_key, 3)))
   {
@@ -394,18 +361,18 @@ create_table_mon:
     goto error;
   }
 
-  /* create one `SPIDER_TABLE_MON' per row in
-  mysql.spider_link_mon_servers with matching db name, table name and
-  link_idx, and add it to `table_mon_list'. */
   do {
     if (!(table_mon = (SPIDER_TABLE_MON *)
       spider_bulk_malloc(spider_current_trx, SPD_MID_GET_PING_TABLE_MON_1, MYF(MY_WME | MY_ZEROFILL),
-        &table_mon, sizeof(SPIDER_TABLE_MON),
-        &tmp_share, sizeof(SPIDER_SHARE),
-        &tmp_connect_info, sizeof(char *) * SPIDER_TMP_SHARE_CHAR_PTR_COUNT,
-        &tmp_connect_info_length, sizeof(uint) * SPIDER_TMP_SHARE_UINT_COUNT,
-        &tmp_long, sizeof(long) * SPIDER_TMP_SHARE_LONG_COUNT,
-        &tmp_longlong, sizeof(longlong) * SPIDER_TMP_SHARE_LONGLONG_COUNT,
+        &table_mon, (uint) (sizeof(SPIDER_TABLE_MON)),
+        &tmp_share, (uint) (sizeof(SPIDER_SHARE)),
+        &tmp_connect_info,
+          (uint) (sizeof(char *) * SPIDER_TMP_SHARE_CHAR_PTR_COUNT),
+        &tmp_connect_info_length,
+          (uint) (sizeof(uint) * SPIDER_TMP_SHARE_UINT_COUNT),
+        &tmp_long, (uint) (sizeof(long) * SPIDER_TMP_SHARE_LONG_COUNT),
+        &tmp_longlong,
+          (uint) (sizeof(longlong) * SPIDER_TMP_SHARE_LONGLONG_COUNT),
         NullS))
     ) {
       spider_sys_index_end(table_link_mon);
@@ -427,7 +394,7 @@ create_table_mon:
       (error_num = spider_get_sys_link_mon_server_id(
         table_link_mon, &table_mon->server_id, mem_root)) ||
       (error_num = spider_get_sys_link_mon_connect_info(
-        table_link_mon, tmp_share, mem_root))
+        table_link_mon, tmp_share, 0, mem_root))
     ) {
       table_link_mon->file->print_error(error_num, MYF(0));
       spider_sys_index_end(table_link_mon);
@@ -491,26 +458,20 @@ error:
   DBUG_RETURN(error_num);
 }
 
-/*
-  creates and return table_mon_list associated with table with `name'
-  and `link_idx'th link.
-*/
 SPIDER_TABLE_MON_LIST *spider_get_ping_table_tgt(
   THD *thd,
-  char *name,                   /* The table name, usually fully qualified */
+  char *name,
   uint name_length,
   int link_idx,
   char *static_link_id,
   uint static_link_id_length,
-  uint32 server_id,             /* The server_id will determine the
-                                `current' field of the returned
-                                `SPIDER_TABLE_MON_LIST'. */
-  spider_string *str,           /* str->c_ptr() == name */
+  uint32 server_id,
+  spider_string *str,
   bool need_lock,
   int *error_num
 ) {
   TABLE *table_tables = NULL;
-  Open_tables_backup open_tables_backup;
+  SPIDER_Open_tables_backup open_tables_backup;
   char table_key[MAX_KEY_LENGTH];
 
   SPIDER_TABLE_MON_LIST *table_mon_list = NULL;
@@ -526,13 +487,17 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_tgt(
   SPD_INIT_ALLOC_ROOT(&mem_root, 4096, 0, MYF(MY_WME));
   if (!(table_mon_list = (SPIDER_TABLE_MON_LIST *)
     spider_bulk_malloc(spider_current_trx, SPD_MID_GET_PING_TABLE_TGT_1, MYF(MY_WME | MY_ZEROFILL),
-      &table_mon_list, sizeof(SPIDER_TABLE_MON_LIST),
-      &tmp_share, sizeof(SPIDER_SHARE),
-      &tmp_connect_info, sizeof(char *) * SPIDER_TMP_SHARE_CHAR_PTR_COUNT,
-      &tmp_connect_info_length, sizeof(uint) * SPIDER_TMP_SHARE_UINT_COUNT,
-      &tmp_long, sizeof(long) * SPIDER_TMP_SHARE_LONG_COUNT,
-      &tmp_longlong, sizeof(longlong) * SPIDER_TMP_SHARE_LONGLONG_COUNT,
-      &key_str, str->length() + 1,
+      &table_mon_list, (uint) (sizeof(SPIDER_TABLE_MON_LIST)),
+      &tmp_share, (uint) (sizeof(SPIDER_SHARE)),
+      &tmp_connect_info,
+        (uint) (sizeof(char *) * SPIDER_TMP_SHARE_CHAR_PTR_COUNT),
+      &tmp_connect_info_length,
+        (uint) (sizeof(uint) * SPIDER_TMP_SHARE_UINT_COUNT),
+      &tmp_long,
+        (uint) (sizeof(long) * SPIDER_TMP_SHARE_LONG_COUNT),
+      &tmp_longlong,
+        (uint) (sizeof(longlong) * SPIDER_TMP_SHARE_LONGLONG_COUNT),
+      &key_str, (uint) (str->length() + 1),
       NullS))
   ) {
     my_error(HA_ERR_OUT_OF_MEM, MYF(0));
@@ -546,7 +511,6 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_tgt(
   memcpy(key_str, str->ptr(), table_mon_list->key_length);
   tmp_share->access_charset = thd->variables.character_set_client;
 
-  /* Open mysql.spider_tables */
   if (
     !(table_tables = spider_open_sys_table(
       thd, SPIDER_SYS_TABLES_TABLE_NAME_STR,
@@ -556,8 +520,6 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_tgt(
     my_error(*error_num, MYF(0));
     goto error;
   }
-  /* store db and table names and link idx in mysql.spider_tables for
-  reading */
   spider_store_tables_name(table_tables, name, name_length);
   if (static_link_id)
   {
@@ -581,10 +543,9 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_tgt(
       goto error;
     }
   }
-  /* Populate tmp_share with info read from mysql.spider_tables */
   if (
     (*error_num = spider_get_sys_tables_connect_info(
-      table_tables, tmp_share, &mem_root)) ||
+      table_tables, tmp_share, 0, &mem_root)) ||
     (*error_num = spider_get_sys_tables_link_status(
       table_tables, tmp_share, 0, &mem_root))
   ) {
@@ -608,8 +569,9 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_tgt(
       tmp_share, name, name_length
     )) ||
     (*error_num = spider_create_conn_keys(tmp_share)) ||
-    /* Pinally, populate `table_mon_list' with newly created
-    `SPIDER_TABLE_MON's */
+/*
+    (*error_num = spider_db_create_table_names_str(tmp_share)) ||
+*/
     (*error_num = spider_get_ping_table_mon(
       thd, table_mon_list, name, name_length, link_idx, server_id, &mem_root,
       need_lock))
@@ -619,26 +581,43 @@ SPIDER_TABLE_MON_LIST *spider_get_ping_table_tgt(
   if (tmp_share->link_statuses[0] == SPIDER_LINK_STATUS_NG)
     table_mon_list->mon_status = SPIDER_LINK_MON_NG;
 
+#if MYSQL_VERSION_ID < 50500
+  if (pthread_mutex_init(&table_mon_list->caller_mutex, MY_MUTEX_INIT_FAST))
+#else
   if (mysql_mutex_init(spd_key_mutex_mon_list_caller,
     &table_mon_list->caller_mutex, MY_MUTEX_INIT_FAST))
+#endif
   {
     *error_num = HA_ERR_OUT_OF_MEM;
     goto error_caller_mutex_init;
   }
+#if MYSQL_VERSION_ID < 50500
+  if (pthread_mutex_init(&table_mon_list->receptor_mutex, MY_MUTEX_INIT_FAST))
+#else
   if (mysql_mutex_init(spd_key_mutex_mon_list_receptor,
     &table_mon_list->receptor_mutex, MY_MUTEX_INIT_FAST))
+#endif
   {
     *error_num = HA_ERR_OUT_OF_MEM;
     goto error_receptor_mutex_init;
   }
+#if MYSQL_VERSION_ID < 50500
+  if (pthread_mutex_init(&table_mon_list->monitor_mutex, MY_MUTEX_INIT_FAST))
+#else
   if (mysql_mutex_init(spd_key_mutex_mon_list_monitor,
     &table_mon_list->monitor_mutex, MY_MUTEX_INIT_FAST))
+#endif
   {
     *error_num = HA_ERR_OUT_OF_MEM;
     goto error_monitor_mutex_init;
   }
+#if MYSQL_VERSION_ID < 50500
+  if (pthread_mutex_init(&table_mon_list->update_status_mutex,
+    MY_MUTEX_INIT_FAST))
+#else
   if (mysql_mutex_init(spd_key_mutex_mon_list_update_status,
     &table_mon_list->update_status_mutex, MY_MUTEX_INIT_FAST))
+#endif
   {
     *error_num = HA_ERR_OUT_OF_MEM;
     goto error_update_status_mutex_init;
@@ -711,8 +690,10 @@ int spider_get_ping_table_gtid_pos(
   int error_num, source_link_idx, need_mon;
   char table_key[MAX_KEY_LENGTH];
   TABLE *table_tables, *table_gtid_pos;
-  Open_tables_backup open_tables_backup_tables;
-  Open_tables_backup open_tables_backup_gtid_pos;
+  SPIDER_Open_tables_backup open_tables_backup_tables;
+#ifdef SPIDER_REQUIRE_DEFINE_FOR_SECONDARY_OPEN_TABLES_BACKUP
+  SPIDER_Open_tables_backup open_tables_backup_gtid_pos;
+#endif
   MEM_ROOT mem_root;
   long link_status;
   long monitoring_binlog_pos_at_failing;
@@ -729,6 +710,7 @@ int spider_get_ping_table_gtid_pos(
       db_name = setted db_name and
       table_name = setted table_name
   */
+#ifdef SPIDER_REQUIRE_DEFINE_FOR_SECONDARY_OPEN_TABLES_BACKUP
   if (
     !(table_tables = spider_open_sys_table(
       thd, SPIDER_SYS_TABLES_TABLE_NAME_STR,
@@ -744,6 +726,45 @@ int spider_get_ping_table_gtid_pos(
       &open_tables_backup_gtid_pos, need_lock, &error_num))
   )
     goto error_open_table_gtid_pos;
+#else
+  TABLE_LIST tables_tables;
+  TABLE_LIST tables_gtid_pos;
+  TABLE_LIST *tables = &tables_tables;
+  LEX_CSTRING db_name =
+  {
+    "mysql",
+    sizeof("mysql") - 1
+  };
+  LEX_CSTRING tbl_name_tables =
+  {
+    SPIDER_SYS_TABLES_TABLE_NAME_STR,
+    SPIDER_SYS_TABLES_TABLE_NAME_LEN
+  };
+  LEX_CSTRING tbl_name_gtid_pos =
+  {
+    SPIDER_SYS_POS_FOR_RECOVERY_TABLE_NAME_STR,
+    SPIDER_SYS_POS_FOR_RECOVERY_TABLE_NAME_LEN
+  };
+  tables_tables.init_one_table(&db_name, &tbl_name_tables, 0, TL_READ);
+  tables_gtid_pos.init_one_table(&db_name, &tbl_name_gtid_pos, 0, TL_READ);
+  MDL_REQUEST_INIT(&tables_tables.mdl_request, MDL_key::TABLE,
+    SPIDER_TABLE_LIST_db_str(&tables_tables),
+    SPIDER_TABLE_LIST_table_name_str(&tables_tables),
+    MDL_SHARED_READ, MDL_TRANSACTION);
+  MDL_REQUEST_INIT(&tables_gtid_pos.mdl_request, MDL_key::TABLE,
+    SPIDER_TABLE_LIST_db_str(&tables_gtid_pos),
+    SPIDER_TABLE_LIST_table_name_str(&tables_gtid_pos),
+    MDL_SHARED_READ, MDL_TRANSACTION);
+  tables_tables.next_global = &tables_gtid_pos;
+  if (spider_sys_open_and_lock_tables(thd, &tables,
+    &open_tables_backup_tables))
+  {
+    error_num = my_errno;
+    goto error_open_table_tables;
+  }
+  table_tables = tables_tables.table;
+  table_gtid_pos = tables_gtid_pos.table;
+#endif
 
   table_tables->use_all_columns();
   table_gtid_pos->use_all_columns();
@@ -830,8 +851,10 @@ int spider_get_ping_table_gtid_pos(
   {
     goto error_sys_index_end;
   }
-  spider_close_sys_table(thd, table_gtid_pos, &open_tables_backup_gtid_pos,
-    need_lock);
+#ifdef SPIDER_REQUIRE_DEFINE_FOR_SECONDARY_OPEN_TABLES_BACKUP
+  spider_close_sys_table(thd, table_gtid_pos,
+    &open_tables_backup_gtid_pos, need_lock);
+#endif
   spider_close_sys_table(thd, table_tables, &open_tables_backup_tables,
     need_lock);
 
@@ -843,28 +866,27 @@ error_get_sys_tables_link_status:
   spider_sys_index_end(table_tables);
 error_sys_index_end:
 error_get_sys_table_by_idx:
-  spider_close_sys_table(thd, table_gtid_pos, &open_tables_backup_gtid_pos,
+#ifdef SPIDER_REQUIRE_DEFINE_FOR_SECONDARY_OPEN_TABLES_BACKUP
+  spider_close_sys_table(thd, table_gtid_pos,
+    &open_tables_backup_gtid_pos,
     need_lock);
 error_open_table_gtid_pos:
+#endif
   spider_close_sys_table(thd, table_tables, &open_tables_backup_tables,
     need_lock);
 error_open_table_tables:
   DBUG_RETURN(error_num);
 }
 
-/*
-  Initialise `spider_mon_table_cache' by scanning the
-  mysql.spider_link_mon_servers table, creating distinct
-  `SPIDER_MON_KEY's with the info and inserting them into the cache.
-*/
 int spider_init_ping_table_mon_cache(
   THD *thd,
   MEM_ROOT *mem_root,
   bool need_lock
 ) {
   int error_num, same;
+  uint old_elements;
   TABLE *table_link_mon = NULL;
-  Open_tables_backup open_tables_backup;
+  SPIDER_Open_tables_backup open_tables_backup;
   SPIDER_MON_KEY mon_key;
   DBUG_ENTER("spider_init_ping_table_mon_cache");
 
@@ -884,7 +906,6 @@ int spider_init_ping_table_mon_cache(
     /* reset */
     spider_mon_table_cache.elements = 0;
 
-    /* start at the first row */
     if ((error_num = spider_sys_index_first(table_link_mon,
       table_link_mon->s->primary_key)))
     {
@@ -901,25 +922,31 @@ int spider_init_ping_table_mon_cache(
       mon_key.table_name_length = SPIDER_SYS_LINK_MON_TABLE_TABLE_NAME_SIZE + 1;
       mon_key.link_id_length = SPIDER_SYS_LINK_MON_TABLE_LINK_ID_SIZE + 1;
       do {
-        /* update content of `mon_key' */
         if ((error_num = spider_get_sys_link_mon_key(table_link_mon, &mon_key,
           mem_root, &same)))
           goto error_get_sys_link_mon_key;
 
-        /* `mon_key' has changed content. since
-        mysql.spider_link_mon_servers is indexed by db_name,
-        table_name, link_idx, and server_id, it is possible that
-        different server_ids share the same mon_key which only has
-        db_name, table_name, link_idx */
         if (!same)
         {
           mon_key.sort = spider_calc_for_sort(3, mon_key.db_name,
             mon_key.table_name, mon_key.link_id);
+          old_elements = spider_mon_table_cache.max_element;
           if (push_dynamic(&spider_mon_table_cache, (uchar *) &mon_key))
           {
             error_num = HA_ERR_OUT_OF_MEM;
             goto error_push_dynamic;
           }
+          if (spider_mon_table_cache.max_element != old_elements)
+          {
+            spider_free_mem_calc(spider_current_trx,
+              spider_mon_table_cache_id,
+              old_elements *
+              spider_mon_table_cache.size_of_element);
+            spider_alloc_calc_mem(spider_current_trx,
+              spider_mon_table_cache,
+              spider_mon_table_cache.max_element *
+              spider_mon_table_cache.size_of_element);
+          }
         }
 
         if ((error_num = spider_sys_index_next(table_link_mon)))
@@ -939,12 +966,16 @@ int spider_init_ping_table_mon_cache(
       (uchar *) dynamic_element(&spider_mon_table_cache, 0, SPIDER_MON_KEY *),
       spider_mon_table_cache.elements, sizeof(SPIDER_MON_KEY),
       (qsort_cmp) spider_compare_for_sort);
-    uint old_elements = spider_mon_table_cache.max_element;
+    old_elements = spider_mon_table_cache.max_element;
     freeze_size(&spider_mon_table_cache);
-    if (spider_mon_table_cache.max_element < old_elements)
+    if (spider_mon_table_cache.max_element != old_elements)
     {
       spider_free_mem_calc(spider_current_trx,
         spider_mon_table_cache_id,
+        old_elements *
+        spider_mon_table_cache.size_of_element);
+      spider_alloc_calc_mem(spider_current_trx,
+        spider_mon_table_cache,
         spider_mon_table_cache.max_element *
         spider_mon_table_cache.size_of_element);
     }
@@ -965,13 +996,6 @@ error_open_sys_table:
   DBUG_RETURN(error_num);
 }
 
-/*
-  Read from msyql.spider_link_mon_servers table fields the db name,
-  table name and link_id and search for them with wild card matching
-  in `spider_mon_table_cache'. store the db name, table name, and
-  link_id of the matching `SPIDER_MON_KEY' back to the table field on
-  success.
-*/
 int spider_ping_table_cache_compare(
   TABLE *table,
   MEM_ROOT *mem_root
@@ -1062,14 +1086,23 @@ long long spider_ping_table_body(
   conv_name.init_calc_mem(SPD_MID_PING_TABLE_BODY_1);
   tmp_str.init_calc_mem(SPD_MID_PING_TABLE_BODY_2);
   conv_name.length(0);
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100002
   server_id = global_system_variables.server_id;
+#else
+  server_id = thd->server_id;
+#endif
   if (
     thd->open_tables != 0 ||
     thd->handler_tables_hash.records != 0 ||
     thd->derived_tables != 0 ||
     thd->lock != 0 ||
+#if MYSQL_VERSION_ID < 50500
+    thd->locked_tables != 0 ||
+    thd->prelocked_mode != NON_PRELOCKED
+#else
     thd->locked_tables_list.locked_tables() ||
     thd->locked_tables_mode != LTM_NONE
+#endif
   ) {
     if (thd->open_tables != 0)
     {
@@ -1092,6 +1125,18 @@ long long spider_ping_table_body(
       my_printf_error(ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_NUM,
         ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_STR_WITH_PTR, MYF(0),
         "thd->lock", thd->lock);
+#if MYSQL_VERSION_ID < 50500
+    } else if (thd->locked_tables != 0)
+    {
+      my_printf_error(ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_NUM,
+        ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_STR_WITH_PTR, MYF(0),
+        "thd->locked_tables", thd->locked_tables);
+    } else if (thd->prelocked_mode != NON_PRELOCKED)
+    {
+      my_printf_error(ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_NUM,
+        ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_STR_WITH_NUM, MYF(0),
+        "thd->prelocked_mode", (longlong) thd->prelocked_mode);
+#else
     } else if (thd->locked_tables_list.locked_tables())
     {
       my_printf_error(ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_NUM,
@@ -1103,6 +1148,7 @@ long long spider_ping_table_body(
       my_printf_error(ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_NUM,
         ER_SPIDER_UDF_CANT_USE_IF_OPEN_TABLE_STR_WITH_NUM, MYF(0),
         "thd->locked_tables_mode", (longlong) thd->locked_tables_mode);
+#endif
     }
     goto error;
   }
@@ -1248,6 +1294,9 @@ long long spider_ping_table_body(
         DBUG_PRINT("info",("spider mon_table_result->result_status=SPIDER_LINK_MON_NG 2"));
         if (table_mon_list->mon_status != SPIDER_LINK_MON_NG)
         {
+/*
+          pthread_mutex_lock(&table_mon_list->update_status_mutex);
+*/
           pthread_mutex_lock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
           if (table_mon_list->mon_status != SPIDER_LINK_MON_NG)
           {
@@ -1262,6 +1311,9 @@ long long spider_ping_table_body(
               conv_name.c_ptr(), conv_name_length, link_idx, TRUE);
             status_changed_to_ng = TRUE;
           }
+/*
+          pthread_mutex_unlock(&table_mon_list->update_status_mutex);
+*/
           pthread_mutex_unlock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
           if (status_changed_to_ng)
           {
@@ -1315,6 +1367,9 @@ long long spider_ping_table_body(
           DBUG_PRINT("info",("spider mon_table_result->result_status=SPIDER_LINK_MON_NG 3"));
           if (table_mon_list->mon_status != SPIDER_LINK_MON_NG)
           {
+/*
+            pthread_mutex_lock(&table_mon_list->update_status_mutex);
+*/
             pthread_mutex_lock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
             if (table_mon_list->mon_status != SPIDER_LINK_MON_NG)
             {
@@ -1329,6 +1384,9 @@ long long spider_ping_table_body(
                 conv_name.c_ptr(), conv_name_length, link_idx, TRUE);
               status_changed_to_ng = TRUE;
             }
+/*
+            pthread_mutex_unlock(&table_mon_list->update_status_mutex);
+*/
             pthread_mutex_unlock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
             if (status_changed_to_ng)
             {
@@ -1372,6 +1430,9 @@ long long spider_ping_table_body(
             mon_table_result->result_status == SPIDER_LINK_MON_NG &&
             table_mon_list->mon_status != SPIDER_LINK_MON_NG
           ) {
+/*
+            pthread_mutex_lock(&table_mon_list->update_status_mutex);
+*/
             pthread_mutex_lock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
             if (table_mon_list->mon_status != SPIDER_LINK_MON_NG)
             {
@@ -1386,6 +1447,9 @@ long long spider_ping_table_body(
                 conv_name.c_ptr(), conv_name_length, link_idx, TRUE);
               status_changed_to_ng = TRUE;
             }
+/*
+            pthread_mutex_unlock(&table_mon_list->update_status_mutex);
+*/
             pthread_mutex_unlock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
             if (status_changed_to_ng)
             {
@@ -1556,9 +1620,9 @@ int spider_ping_table_mon_from_table(
   SPIDER_SHARE *share,
   int base_link_idx,
   uint32 server_id,
-  char *conv_name,              /* Usually fully qualified table name */
+  char *conv_name,
   uint conv_name_length,
-  int link_idx,                 /* The link id to ping */
+  int link_idx,
   char *where_clause,
   uint where_clause_length,
   long monitoring_kind,
@@ -1568,6 +1632,9 @@ int spider_ping_table_mon_from_table(
 ) {
   int error_num = 0, current_mon_count, flags;
   uint32 first_sid;
+/*
+  THD *thd = trx->thd;
+*/
   SPIDER_TABLE_MON_LIST *table_mon_list;
   SPIDER_TABLE_MON *table_mon;
   SPIDER_MON_TABLE_RESULT mon_table_result;
@@ -1636,7 +1703,6 @@ int spider_ping_table_mon_from_table(
   if (monitoring_flag & 1)
     flags |= SPIDER_UDF_PING_TABLE_USE_ALL_MONITORING_NODES;
 
-  /* Get or create `table_mon_list' for `conv_name_str'. */
   if (!(table_mon_list = spider_get_ping_table_mon_list(trx, thd,
     &conv_name_str, conv_name_length, link_idx,
     share->static_link_ids[link_idx],
@@ -1668,8 +1734,6 @@ int spider_ping_table_mon_from_table(
     table_mon = table_mon_list->current;
     first_sid = table_mon->server_id;
     current_mon_count = 1;
-    /* Call spider_ping_table on each table_mon of `table_mon_list',
-    until one succeeds */
     while (TRUE)
     {
       DBUG_PRINT("info",("spider thd->killed=%s",
@@ -1714,13 +1778,16 @@ int spider_ping_table_mon_from_table(
           if (!spider_db_udf_ping_table_mon_next(
             thd, table_mon, mon_conn, &mon_table_result, conv_name,
             conv_name_length, link_idx,
-            where_clause, where_clause_length, /*first_sid=*/-1, table_mon_list->list_size,
+            where_clause, where_clause_length, -1, table_mon_list->list_size,
             0, 0, 0, flags, monitoring_limit))
           {
             if (
               mon_table_result.result_status == SPIDER_LINK_MON_NG &&
               table_mon_list->mon_status != SPIDER_LINK_MON_NG
             ) {
+/*
+              pthread_mutex_lock(&table_mon_list->update_status_mutex);
+*/
               pthread_mutex_lock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
               if (table_mon_list->mon_status != SPIDER_LINK_MON_NG)
               {
@@ -1735,6 +1802,9 @@ int spider_ping_table_mon_from_table(
                 spider_sys_log_tables_link_failed(thd, conv_name,
                   conv_name_length, link_idx, need_lock);
               }
+/*
+              pthread_mutex_unlock(&table_mon_list->update_status_mutex);
+*/
               pthread_mutex_unlock(&spider_udf_table_mon_mutexes[table_mon_list->mutex_hash]);
             }
             table_mon_list->last_caller_result = mon_table_result.result_status;
